# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Flow API routes."""
# mypy: disable-error-code="unused-ignore,assignment"
# pylint: disable=broad-except,wrong-import-order,ungrouped-imports
# pyright: reportUnreachable=false,reportCallInDefaultInitializer=false

import json
import logging
from pathlib import Path
from typing import Any

try:
    from typing import Literal  # noqa
except ImportError:
    from typing_extensions import Literal  # noqa

from fastapi import APIRouter, Depends, HTTPException, Query
from waldiez.exporter import WaldiezExporter

from waldiez_studio.models import PathItem, SaveFlowRequest
from waldiez_studio.utils.sync import sync_to_async

from .common import check_flow_path, get_root_directory

api = APIRouter()
LOG = logging.getLogger(__name__)


def export_flow_sync(
    flow_path: Path,
    extension: Literal["py", "ipynb"],
    root_dir: Path,
) -> tuple[bool, int, str]:
    """Export flow synchronously.

    Parameters
    ----------
    flow_path : Path
        The path to the flow.
    extension : Literal["py", "ipynb"]
        The extension to export to.
    root_dir : Path
        The root directory of the workspace.

    Returns
    -------
    tuple[bool, int str]
        Whether the export was successful and the destination or
        an error message with the status code.
    """
    if extension not in ("py", "ipynb"):
        LOG.error("Invalid extension: %s", extension)
        return False, 422, "Invalid extension"
    try:
        exporter = WaldiezExporter.load(file_path=flow_path)
    except BaseException as e:
        LOG.error("Error loading flow: %s", e)
        return False, 400, str(e)
    destination = flow_path.with_suffix(f".{extension}")
    try:
        exporter.export(destination, force=True)
    except BaseException as e:
        LOG.error("Error exporting flow: %s", e)
        return False, 500, str(e)
    return True, 200, str(destination.relative_to(root_dir))


@api.get(
    "/flow",
    responses={
        200: {"description": "The content of the flow."},
        400: {"description": "Error: Invalid path or file type."},
        404: {"description": "Error: File does not exist."},
        500: {"description": "Error: Could not read the flow."},
    },
)
async def get_flow_contents(
    flow_path: Path = Depends(check_flow_path),
) -> dict[str, Any]:
    """Get the contents of a flow.

    Parameters
    ----------
    flow_path : Path
        The path to the flow.

    Returns
    -------
    WaldiezFlow
        The content of the flow.

    Raises
    ------
    HTTPException
        If the path is invalid, the file does not exist,
        the file is not a file, the file type is invalid,
        or an error occurs while reading.
    """
    return await sync_to_async(get_flow_contents_sync)(flow_path)


@api.post(
    "/flow",
    responses={
        200: {"description": "Flow saved successfully."},
        400: {"description": "Error: Invalid path or file type."},
        404: {"description": "Error: File does not exist."},
        500: {"description": "Error: Could not save the flow."},
    },
)
async def save_flow_contents(
    data: SaveFlowRequest,
    flow_path: Path = Depends(check_flow_path),
) -> PathItem:
    """Save the content of a flow.

    Parameters
    ----------
    flow_path : Path
        The path to the flow.
    data : dict[str, Any]
        The content of the flow.

    Returns
    -------
    dict[str, str]
        Whether the save was successful or an error message.

    Raises
    ------
    HTTPException
        If the path is invalid, the file type is invalid,
        or an error occurs while writing.
    """
    contents = (
        data.contents
        if isinstance(data.contents, str)
        else json.dumps(data.contents)
    )
    try:
        await sync_to_async(flow_path.write_text)(contents, encoding="utf-8")
        LOG.debug("Saved flow at %s", flow_path)
    except Exception as e:
        LOG.error("Error writing file: %s", e)
        raise HTTPException(
            status_code=500, detail="Error: Could not save the flow"
        ) from e
    return PathItem(
        path=str(flow_path.relative_to(flow_path.parent.parent)),
        type="file",
        name=flow_path.name,
    )


@api.post(
    "/flow/export",
    responses={
        200: {"description": "The content of the flow."},
        400: {"description": "Error: Invalid path or file type."},
        404: {"description": "Error: File does not exist."},
        500: {"description": "Error: Could not export the flow."},
    },
)
async def export_flow(
    extension: Literal["py", "ipynb"] = Query(
        ..., description="The extension to export to."
    ),
    flow_path: Path = Depends(check_flow_path),
    root_dir: Path = Depends(get_root_directory),
) -> PathItem:
    """Export a flow to a Python script or Jupyter notebook.

    Parameters
    ----------
    extension : Literal["py", "ipynb"]
        The extension to export to.
    flow_path : Path
        The path to the flow.
    root_dir : Path
        The root directory of the workspace.

    Returns
    -------
    dict[str, str]
        Whether the export was successful and the destination or error message.

    Raises
    ------
    HTTPException
        If the path is invalid, the file type is invalid,
        or an error occurs while exporting.
    """
    success, status, message = await sync_to_async(export_flow_sync)(
        flow_path, extension, root_dir
    )
    if not success:
        raise HTTPException(status_code=status, detail=message)
    return PathItem(
        path=message,
        type="file",
        name=flow_path.name,
    )


def get_flow_contents_sync(flow_path: Path) -> dict[str, Any]:
    """Get the contents of a flow synchronously.

    Parameters
    ----------
    flow_path : Path
        The path to the flow.

    Returns
    -------
    WaldiezFlow | dict[str, Any]
        The content of the flow.

    Raises
    ------
    HTTPException
        If an error occurs while reading.
    """
    try:
        flow_text = flow_path.read_text(encoding="utf-8")
        LOG.debug("Flow contents: %s", flow_text)
        return json.loads(flow_text)
    except Exception as e:
        LOG.error("Error reading file: %s", e)
        raise HTTPException(
            status_code=500, detail="Error: Could not read the flow contents"
        ) from e



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Factory to create the appropriate engine."""

from pathlib import Path
from typing import Final

from fastapi import WebSocket

from .base import Engine

SUPPORTED_EXTS: Final[set[str]] = {".py", ".ipynb", ".waldiez"}


# pylint: disable=unused-argument
async def make_engine(
    *,
    file_path: Path,
    root_dir: Path,
    websocket: WebSocket,
) -> Engine:
    """
    Create an engine instance for the given file type.

    - .py      -> SubprocessEngine
    - .ipynb   -> NotebookEngine
    - .waldiez -> WaldiezEngine (compile -> subprocess)

    Parameters
    ----------
    file_path : Path
        The path of the file to run.
    root_dir : Path
        The root directory of the file
    websocket : Websocket
        The websocket to use for communication.

    Raises
    ------
    ValueError
        If the file extension is is not supported.
    NotImplementedError
        Well, this should be removed.

    Returns
    -------
    Engine
        The engine instance for the selected file.
    """
    ext = file_path.suffix.lower()
    if ext not in SUPPORTED_EXTS:
        raise ValueError(f"Unsupported extension: {ext}")
    # pylint: disable=import-outside-toplevel
    # Lazy import.
    if ext == ".py":
        from .subprocess_engine import SubprocessEngine

        return SubprocessEngine(
            file_path=file_path, root_dir=root_dir, websocket=websocket
        )

    if ext == ".ipynb":
        from .notebook_engine import NotebookEngine

        return NotebookEngine(
            file_path=file_path, root_dir=root_dir, websocket=websocket
        )

    if ext == ".waldiez":
        from .waldiez_engine import WaldiezEngine

        return WaldiezEngine(
            file_path=file_path, root_dir=root_dir, websocket=websocket
        )
    raise ValueError(f"Unsupported extension: {ext}")  # pragma: no cover



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Notebook engine implementation."""

# pyright: reportUnknownMemberType=false,reportUnknownArgumentType=false
# pylint: disable=broad-exception-caught,too-many-try-statements

from __future__ import annotations

import asyncio
import contextlib
import json
import time
from pathlib import Path
from typing import Any

from fastapi import WebSocket
from jupyter_client.asynchronous.client import AsyncKernelClient

from .base import Engine
from .kernel import KernelManager


class NotebookEngine(Engine):
    """Notebook engine implementation."""

    file_path: Path
    root_dir: Path
    websocket: WebSocket
    max_queue: int = 1000
    exec_timeout_sec: float = 120.0  # per-code-cell timeout

    kc: AsyncKernelClient | None = None
    _monitor_tasks: list[asyncio.Task[Any]] | None = None
    _queue: asyncio.Queue[dict[str, Any]] | None = None
    _start_ts: float = 0.0

    async def start(self, start_msg: dict[str, Any] | None = None) -> None:
        """Start the run (after receiving the initial 'op=start' message).

        Parameters
        ----------
        start_msg : dict[str, Any]
            Optional start arguments to pass.
        """
        start_msg = start_msg or {}
        fresh = bool(start_msg.get("freshKernel", False))

        if fresh:
            await KernelManager.shutdown_kernel(now=True)

        km = await KernelManager.get()
        self.kc = km.client()
        self.kc.start_channels()
        await self.kc.wait_for_ready(timeout=60)

        self._start_ts = time.time()

        self._queue = asyncio.Queue(maxsize=self.max_queue)
        await self._enqueue(
            {"type": "run_status", "data": {"state": "started"}}
        )

        self._monitor_tasks = [
            asyncio.create_task(self._forward_iopub()),
            asyncio.create_task(self._monitor_stdin()),
            asyncio.create_task(self._send_queued()),
        ]

        # Load notebook
        try:
            nb = json.loads(self.file_path.read_text(encoding="utf-8"))
        except Exception as exc:
            await self._enqueue(
                {
                    "type": "run_stderr",
                    "data": {"text": f"Failed to read notebook: {exc}\n"},
                }
            )
            return

        # Run all code cells, serialized by the global lock
        idx = -1
        async with KernelManager.lock():
            for cell in nb.get("cells", []):
                if cell.get("cell_type") != "code":
                    continue
                idx += 1
                code = "".join(cell.get("source", []))
                await self._enqueue(
                    {"type": "cell_start", "data": {"index": idx}}
                )

                try:
                    self.kc.execute(code)  # fire
                    status = await asyncio.wait_for(
                        self._wait_shell_reply(), timeout=self.exec_timeout_sec
                    )
                except asyncio.TimeoutError:
                    await KernelManager.interrupt()
                    await self._enqueue(
                        {
                            "type": "run_stderr",
                            "data": {
                                "text": "Execution timed out; interrupted.\n"
                            },
                        }
                    )
                    await self._enqueue(
                        {
                            "type": "cell_end",
                            "data": {"index": idx, "status": "timeout"},
                        }
                    )
                    break
                except Exception as exc:
                    await self._enqueue(
                        {
                            "type": "run_stderr",
                            "data": {"text": f"Execution error: {exc}\n"},
                        }
                    )
                    await self._enqueue(
                        {
                            "type": "cell_end",
                            "data": {"index": idx, "status": "error"},
                        }
                    )
                    break
                else:
                    await self._enqueue(
                        {
                            "type": "cell_end",
                            "data": {"index": idx, "status": status},
                        }
                    )

    async def handle_client(self, msg: dict[str, Any]) -> None:
        """Handle subsequent client control messages (stdin/interrupt/etc.).

        Parameters
        ----------
        msg : dict[str, Any]
            Message from the client.
        """
        if not self.kc:
            return
        op = msg.get("op")
        if op == "interrupt":
            await KernelManager.interrupt()
        elif op == "restart":
            await KernelManager.restart()
            with contextlib.suppress(Exception):
                self.kc.stop_channels()
            with contextlib.suppress(Exception):
                self.kc.start_channels()
                await self.kc.wait_for_ready(timeout=60)
        elif op == "input_reply":
            with contextlib.suppress(Exception):
                self.kc.input(msg.get("value", ""))

    async def shutdown(self) -> None:
        """Finalize the run and emit 'run_end' if appropriate."""
        # Cancel monitors
        if self._monitor_tasks:
            for task in self._monitor_tasks:
                task.cancel()
            for task in self._monitor_tasks:
                with contextlib.suppress(asyncio.CancelledError, RuntimeError):
                    await task

        if self.kc:
            with contextlib.suppress(Exception):
                self.kc.stop_channels()

        elapsed = int((time.time() - self._start_ts) * 1000)
        await self._send_safe(
            {"type": "run_end", "data": {"status": "ok", "elapsedMs": elapsed}}
        )

        self._monitor_tasks = None
        self._queue = None

    # pylint: disable=too-complex
    async def _forward_iopub(self) -> None:  # noqa: C901
        if self.kc is None:
            return
        try:
            async for m in self.kc.iopub_channel:
                t = m.get("header", {}).get("msg_type")
                c = m.get("content", {})
                if t == "status":
                    await self._enqueue(
                        {
                            "type": "kernel_status",
                            "data": {
                                "execution_state": c.get("execution_state")
                            },
                        }
                    )
                elif t == "stream":
                    kind = "stdout" if c.get("name") == "stdout" else "stderr"
                    await self._enqueue(
                        {
                            "type": f"run_{kind}",
                            "data": {"text": c.get("text", "")},
                        }
                    )
                elif t in {"display_data", "execute_result"}:
                    data = c.get("data", {})
                    if "image/png" in data:
                        await self._enqueue(
                            {
                                "type": "cell_output",
                                "data": {
                                    "mime": "image/png",
                                    "b64": data["image/png"],
                                },
                            }
                        )
                    if "text/plain" in data:
                        await self._enqueue(
                            {
                                "type": "cell_output",
                                "data": {
                                    "mime": "text/plain",
                                    "text": data["text/plain"],
                                },
                            }
                        )
                    if "text/html" in data:
                        html = data["text/html"]
                        if isinstance(html, list):
                            html = "".join(html)
                        await self._enqueue(
                            {
                                "type": "cell_output",
                                "data": {"mime": "text/html", "text": html},
                            }
                        )
                elif t == "error":
                    tb = "\n".join(c.get("traceback", []))
                    await self._enqueue(
                        {"type": "run_stderr", "data": {"text": tb}}
                    )
        except asyncio.CancelledError as exc:
            raise asyncio.CancelledError("Task was cancelled") from exc
        except Exception:
            pass

    async def _monitor_stdin(self) -> None:
        if self.kc is None:
            return
        try:
            while True:
                try:
                    msg = await asyncio.wait_for(
                        self.kc.stdin_channel.get_msg(), timeout=1.0
                    )
                except asyncio.TimeoutError:
                    continue
                except Exception:
                    break
                if msg.get("header", {}).get("msg_type") == "input_request":
                    c = msg.get("content", {})
                    await self._enqueue(
                        {
                            "type": "input_request",
                            "data": {
                                "prompt": c.get("prompt", ""),
                                "password": bool(c.get("password", False)),
                            },
                        }
                    )
        except asyncio.CancelledError as exc:
            raise asyncio.CancelledError("Task was cancelled") from exc
        except Exception:
            pass

    async def _wait_shell_reply(self) -> str:
        if self.kc is None:
            return "error"
        while True:
            msg = await self.kc.shell_channel.get_msg()
            status = msg.get("content", {}).get("status")
            if status in {"ok", "error", "aborted"}:
                return status

    async def _enqueue(self, message: dict[str, Any]) -> None:
        if self._queue is None:
            return
        try:
            self._queue.put_nowait(message)
        except asyncio.QueueFull:
            with contextlib.suppress(asyncio.QueueEmpty):
                _ = self._queue.get_nowait()
            await self._queue.put(message)

    async def _send_queued(self) -> None:
        if self._queue is None:
            return
        try:
            while True:
                try:
                    msg = await asyncio.wait_for(self._queue.get(), timeout=1.0)
                except asyncio.TimeoutError:
                    continue
                await self._send_safe(msg)
        except asyncio.CancelledError:
            for _ in range(50):
                if self._queue.empty():
                    break
                with contextlib.suppress(Exception):
                    await self._send_safe(self._queue.get_nowait())
            raise

    async def _send_safe(self, message: dict[str, Any]) -> None:
        with contextlib.suppress(Exception):
            await self.websocket.send_json(message)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pyright: reportArgumentType=false,reportUnusedParameter=false

"""Application entry point."""

import logging
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager

from fastapi import FastAPI, Response
from fastapi.middleware.cors import CORSMiddleware
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.middleware.trustedhost import TrustedHostMiddleware
from fastapi.openapi.docs import (
    get_swagger_ui_html,
    get_swagger_ui_oauth2_redirect_html,
)
from fastapi.responses import FileResponse, HTMLResponse, RedirectResponse
from fastapi.staticfiles import StaticFiles
from uvicorn.middleware.proxy_headers import ProxyHeadersMiddleware

from waldiez_studio._version import __version__
from waldiez_studio.config.settings import get_settings
from waldiez_studio.middleware import ExtraHeadersMiddleware
from waldiez_studio.routes import api_router, ws_router, ws_term_router
from waldiez_studio.utils.extra_static import ensure_extra_static_files
from waldiez_studio.utils.paths import get_static_dir

LOG = logging.getLogger(__name__)

settings = get_settings()

STATIC_DIR = get_static_dir()
FRONTEND_DIR = STATIC_DIR / "frontend"
MONACO_DIR = STATIC_DIR / "monaco"
MONACO_VS = MONACO_DIR / "vs"
MONACO_MIN_MAPS = MONACO_DIR / "min-maps"
SWAGGER_DIR = STATIC_DIR / "swagger"
FRONTEND_DIR.mkdir(parents=True, exist_ok=True)
MONACO_DIR.mkdir(parents=True, exist_ok=True)
MONACO_VS.mkdir(parents=True, exist_ok=True)
MONACO_MIN_MAPS.mkdir(parents=True, exist_ok=True)
SWAGGER_DIR.mkdir(parents=True, exist_ok=True)


@asynccontextmanager
async def lifespan(
    _app: FastAPI,
) -> AsyncIterator[None]:
    """Application lifespan context manager.

    Yields
    ------
    None
        The application lifespan

    Raises
    ------
    RuntimeError
        If the required static files could not be ensured
    """
    try:
        await ensure_extra_static_files(STATIC_DIR)
        LOG.debug("Monaco editor files are ready.")
    except BaseException as e:  # pragma: no cover
        LOG.error("Failed to prepare extra static files.")
        raise RuntimeError("Critical setup step failed.") from e
    yield


app = FastAPI(
    lifespan=lifespan,
    docs_url=None,
    redoc_url=None,
    title="Waldiez Studio",
    description="A simple file manager for Waldiez.",
    version=__version__,
    openapi_url="/openapi.json",
)

# noinspection PyUnresolvedReferences
swagger_ui_oauth2_redirect_url = (
    app.swagger_ui_oauth2_redirect_url or "/docs/oauth2-redirect"
)

# middlewares
app.add_middleware(
    CORSMiddleware,
    allow_origins=settings.trusted_origins,
    allow_origin_regex=settings.trusted_origin_regex,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
app.add_middleware(
    ProxyHeadersMiddleware,  # type: ignore[arg-type,unused-ignore]
    trusted_hosts=settings.trusted_hosts,
)
app.add_middleware(
    TrustedHostMiddleware,
    allowed_hosts=settings.trusted_hosts,
    www_redirect=False,
)
app.add_middleware(GZipMiddleware, minimum_size=1000)
app.add_middleware(
    ExtraHeadersMiddleware,
    csp=True,
    exclude_patterns=[
        "^/docs",
        "^/openapi.json",
        "^/favicon.ico",
        "^/robots.txt",
        "^/health",
        "^/healthz",
    ],
    force_ssl=settings.force_ssl,
    max_age=31556926,
)


# self-hosted Swagger UI
@app.get("/docs", include_in_schema=False)
async def custom_swagger_ui_html() -> HTMLResponse:
    """Get the custom Swagger UI HTML page.

    Returns
    -------
    HTMLResponse
        The custom Swagger UI HTML page
    """
    # noinspection PyUnresolvedReferences
    return get_swagger_ui_html(
        openapi_url=app.openapi_url or "/openapi.json",
        title=app.title + " - Swagger UI",
        oauth2_redirect_url=app.swagger_ui_oauth2_redirect_url,
        swagger_js_url="/swagger/js/swagger-ui-bundle.js",
        swagger_css_url="/swagger/css/swagger-ui.css",
        swagger_favicon_url="/frontend/favicon.ico",
    )


@app.get("/docs/", include_in_schema=False)
async def redirect_docs() -> Response:
    """Redirect to the custom Swagger UI HTML page.

    Returns
    -------
    Response
        The redirect to the custom Swagger UI HTML page
    """
    return RedirectResponse(url="/docs")


@app.get(swagger_ui_oauth2_redirect_url, include_in_schema=False)
async def swagger_ui_redirect() -> HTMLResponse:  # pragma: no cover
    """Redirect to the Swagger UI OAuth2 page.

    Returns
    -------
    HTMLResponse
        The Swagger UI OAuth2 page
    """
    return get_swagger_ui_oauth2_redirect_html()


# mount static files
app.mount("/vs", StaticFiles(directory=MONACO_VS), name="vs")
app.mount(
    "/min-maps",
    StaticFiles(directory=MONACO_MIN_MAPS),
    name="min-maps",
)

app.mount(
    "/swagger",
    StaticFiles(directory=SWAGGER_DIR),
    name="swagger",
)

# include api routes
app.include_router(api_router, prefix="/api")
app.include_router(ws_router)
app.include_router(ws_term_router)


# common routes
@app.get("/robots.txt", include_in_schema=False)
async def robots() -> Response:
    """Serve the robots.txt file.

    Returns
    -------
    Response
        The robots.txt file
    """
    return FileResponse(FRONTEND_DIR / "robots.txt")


@app.get("/favicon.ico", include_in_schema=False)
async def favicon() -> Response:
    """Serve the favicon.

    Returns
    -------
    Response
        The favicon file

    """
    return FileResponse(FRONTEND_DIR / "favicon.ico")


@app.get("/health")
@app.get("/healthz")
async def health_check() -> Response:
    """Health check.

    Returns
    -------
    Response
        The health check
    """
    return Response(status_code=200)


# mount frontend static files
app.mount(
    "/frontend",
    StaticFiles(directory=FRONTEND_DIR),
    name="frontend",
)


# pylint: disable=unused-argument
# noinspection PyUnusedLocal
@app.get("/{full_path:path}", include_in_schema=False)
async def catch_all(full_path: str = "") -> Response:
    """Serve the frontend's index.html.

    Parameters
    ----------
    full_path : str
        The full request path

    Returns
    -------
    Response
        The frontend
    """
    return FileResponse(FRONTEND_DIR / "index.html")



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Models for the flow API."""

from typing import Any

from pydantic import BaseModel

# pylint: disable=too-few-public-methods


class SaveFlowRequest(BaseModel):
    """Represents a flow in the workspace."""

    contents: dict[str, Any] | str



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Routes module for Waldiez Studio API."""

from fastapi import APIRouter

from .flow import api as flow_router
from .terminal_ws import router as ws_term_router
from .workspace import api as workspace_router
from .ws import router as ws_router

api_router = APIRouter()

api_router.include_router(workspace_router, tags=["Workspace"])
api_router.include_router(flow_router, tags=["Flow"])

__all__ = ["api_router", "ws_router", "ws_term_router"]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Command line interface module."""

# pylint: disable=missing-param-doc,missing-return-doc,missing-raises-doc
# pylint: disable=invalid-name

# pyright: reportCallInDefaultInitializer=false,reportConstantRedefinition=false

import logging
import logging.config
import os
import sys
from enum import Enum
from pathlib import Path

import typer
import uvicorn

try:
    # noinspection PyUnresolvedReferences
    from dotenv import load_dotenv
except ImportError:
    pass
else:
    load_dotenv()

try:
    from waldiez_studio._version import __version__
except ImportError:
    sys.path.append(str(Path(__file__).parent))
    from waldiez_studio._version import __version__

from waldiez_studio._logging import get_logging_config
from waldiez_studio.config import (
    Settings,
    get_default_domain_name,
    get_default_host,
    get_default_port,
)


class LogLevel(str, Enum):
    """Log level enum."""

    CRITICAL = "CRITICAL"
    ERROR = "ERROR"
    WARNING = "WARNING"
    INFO = "INFO"
    DEBUG = "DEBUG"


DEFAULT_LOG_LEVEL = os.environ.get("WALDIEZ_STUDIO_LOG_LEVEL", "INFO")
LOG_LEVELS = {level.value for level in LogLevel}
if DEFAULT_LOG_LEVEL not in LOG_LEVELS:  # pragma: no cover
    DEFAULT_LOG_LEVEL = "INFO"


app = typer.Typer(
    name="waldiez-studio",
    help="Waldiez Studio",
    context_settings={
        "allow_extra_args": True,
        "ignore_unknown_options": True,
    },
    add_completion=False,
    no_args_is_help=False,
    invoke_without_command=True,
    add_help_option=True,
    pretty_exceptions_short=False,
)


# pylint: disable=too-many-locals,unused-argument
@app.command()
def run(
    host: str = typer.Option(
        default=get_default_host(),
        help="The host to run the server on",
    ),
    port: int = typer.Option(
        default=get_default_port(),
        help="The port to run the server on",
    ),
    reload: bool = typer.Option(
        default=False,
        help="Reload the server on file changes",
    ),
    log_level: LogLevel = typer.Option(
        default=DEFAULT_LOG_LEVEL,
        help="The log level",
        case_sensitive=False,
    ),
    domain_name: str = get_default_domain_name(),
    trusted_hosts: list[str] = typer.Option(
        default_factory=list,
    ),
    trusted_origins: list[str] = typer.Option(
        default_factory=list,
    ),
    force_ssl: bool = typer.Option(
        default=False,
        help="Force SSL",
    ),
    version: bool = typer.Option(
        False,
        "--version",
        help="Show the version",
    ),
) -> None:
    """Get the command line arguments."""
    if version:
        typer.echo(f"Waldiez Studio {__version__}")
        raise typer.Exit()
    logging_config = get_logging_config(log_level.value.upper())
    logging.config.dictConfig(logging_config)
    logger = logging.getLogger("waldiez::studio")
    logger.debug("Starting the application")
    settings = Settings(
        domain_name=domain_name,
        trusted_hosts=trusted_hosts,
        trusted_origins=trusted_origins,
        force_ssl=force_ssl,
        host=host,
        port=port,
    )
    logger.debug("Settings: %s", settings.model_dump_json(indent=2))
    settings.to_env()
    this_dir = Path(__file__).parent
    this_dir_name = this_dir.name
    chdir_to = str(this_dir.parent)
    app_module_path = f"{this_dir_name}.main"
    if host in ("localhost", "127.0.0.1"):  # pragma: no cover
        host = "0.0.0.0"
    uvicorn.run(
        f"{app_module_path}:app",
        host=host,
        port=port,
        reload=reload,
        app_dir=chdir_to,
        date_header=False,
        server_header=False,
        reload_dirs=[str(this_dir)] if reload else None,
        reload_includes=["waldiez_studio/**/*.py"] if reload else None,
        reload_excludes=(
            [
                "**/waldiez_out/**/*",
                ".*",
                ".py[cod]",
                ".sw.*",
                "~*",
                "**/files/**/*",
                ".venv/*",
            ]
            if reload
            else None
        ),
        log_level=log_level.lower(),
        log_config=logging_config,
        proxy_headers=True,
        forwarded_allow_ips="*",
        ws_ping_timeout=None,
        ws="websockets",
        # ws="wsproto",
    )


if __name__ == "__main__":
    app()



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pyright: reportUnknownVariableType=false,reportUnusedParameter=false
# pyright: reportUnknownArgumentType=false

"""Waldiez Studio settings module."""

import os
from typing import Any

from pydantic import ValidationInfo, field_validator
from pydantic_settings import (
    BaseSettings,
    CliSettingsSource,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
)

from .lib import (
    ENV_PREFIX,
    get_default_domain_name,
    get_default_host,
    get_default_port,
    get_trusted_hosts,
    get_trusted_origins,
)


def to_kebab(value: str) -> str:
    """Convert a string to kebab case.

    Parameters
    ----------
    value : str
        The string to convert

    Returns
    -------
    str
        The converted string
    """
    return value.replace("_", "-")


class Settings(BaseSettings):
    """Settings class."""

    host: str = get_default_host()
    port: int = get_default_port()
    domain_name: str = get_default_domain_name()
    force_ssl: bool = False
    trusted_hosts: str | list[str] = get_trusted_hosts(
        domain_name=domain_name, host=host
    )
    trusted_origins: str | list[str] = get_trusted_origins(
        domain_name=domain_name,
        port=port,
        force_ssl=force_ssl,
        host=host,
    )
    trusted_origin_regex: str | None = None

    model_config = SettingsConfigDict(
        alias_generator=to_kebab,
        populate_by_name=True,
        env_prefix=ENV_PREFIX,
        case_sensitive=False,
        extra="ignore",
        cli_parse_args=True,
        cli_ignore_unknown_args=True,
        cli_implicit_flags=True,
        cli_prefix="",
    )

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        """Customize the sources.

        Parameters
        ----------
        settings_cls : Type[BaseSettings]
            The settings class
        init_settings : PydanticBaseSettingsSource
            The init settings
        env_settings : PydanticBaseSettingsSource
            The environment settings
        dotenv_settings : PydanticBaseSettingsSource
            The dotenv settings
        file_secret_settings : PydanticBaseSettingsSource
            The file secret settings
        Returns
        -------
        Tuple[PydanticBaseSettingsSource, ...]
            The sources to use priority-wise
        """
        return (
            CliSettingsSource(
                settings_cls,
                cli_implicit_flags=True,
                cli_ignore_unknown_args=True,
                cli_prefix="",
                cli_parse_args=True,
            ),
            env_settings,
            init_settings,
            dotenv_settings,
            file_secret_settings,
        )

    def to_env(self) -> None:
        """Set the environment variables."""
        for key, value in self.model_dump().items():
            if value:
                env_key = f"{ENV_PREFIX}{key.upper()}"
                env_value = (
                    str(value)
                    if not isinstance(value, list)
                    else ",".join(value)
                )
                os.environ[env_key] = env_value

    # pylint: disable=unused-argument
    # noinspection PyUnusedLocal,PyNestedDecorators
    @field_validator("trusted_hosts", "trusted_origins", mode="before")
    @classmethod
    def split_value(cls, value: Any, info: ValidationInfo) -> list[str]:
        """Split the value if it is a string.

        Parameters
        ----------
        value : Any
            The value
        info : ValidationInfo
            The validation info

        Returns
        -------
        list[str]
            The value as a list
        """
        if isinstance(value, str):
            if value.count(",") >= 1:
                return value.split(",")

            return [value]
        return value


def get_settings() -> Settings:
    """Get the settings.

    Returns
    -------
    Settings
        The settings
    """
    settings = Settings()
    is_testing = os.environ.get("WALDIEZ_STUDIO_TESTING", "False") == "true"
    if is_testing:  # pragma: no branch
        settings.trusted_hosts = ["test"]
        settings.trusted_origins = ["http://test"]
    return settings



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Utilities for managing extra static files."""
# pylint: disable=broad-except,too-many-try-statements,invalid-name

import asyncio
import hashlib
import io
import json
import logging
import shutil
import sys
import tarfile
from datetime import datetime, timedelta, timezone
from pathlib import Path

import httpx
from packaging import version

LOG = logging.getLogger(__name__)
HERE = Path(__file__).parent
ROOT_DIR = HERE.parent
STATIC_ROOT = ROOT_DIR / "static"

REGISTRY_BASE_URL = "https://registry.npmjs.org"
PACKAGE_NAME = "monaco-editor"

# 0.53.0 does not seem to play well with @monaco-editor/react
# let's check periodically and change it to None when we are good.
PINNED_VERSION: str | None = "0.54.0"
SWAGGER_DIST = (
    "https://raw.githubusercontent.com/swagger-api/swagger-ui/master/dist"
)
SWAGGER_FILEs = [
    "swagger-ui.css",
    "swagger-ui.css.map",
    "swagger-ui-bundle.js",
    "swagger-ui-bundle.js.map",
]
SWAGGER_ASSETS = [
    (
        file,
        "css" if file.endswith((".css", "css.map")) else "js",
        f"{SWAGGER_DIST}/{file}",
    )
    for file in SWAGGER_FILEs
]


class DownloadError(Exception):
    """Custom exception for download errors."""


def check_cached_details(file_path: Path) -> tuple[str, str, str] | None:
    """Check if the cached package details are recent.

    Parameters
    ----------
    file_path : Path
        The path to the file.

    Returns
    -------
    tuple[str, str, str] | None
        The cached package details if they are recent, None otherwise.
    """
    if not file_path.exists():
        return None
    # pylint: disable=broad-except
    try:
        data = json.loads(file_path.read_text(encoding="utf-8"))
    except BaseException as e:
        LOG.error("Error checking cached details: %s", e)
        return None
    try:
        last_check = datetime.fromisoformat(data.get("last_check", ""))
        vs_version = data.get("version")
        url = data.get("url")
        sha_sum = data.get("sha_sum")
    except BaseException as e:
        LOG.error("Error parsing last check timestamp: %s", e)
        return None
    if not all((last_check, vs_version, url, sha_sum)):
        return None
    if datetime.now(timezone.utc) - last_check < timedelta(days=1):
        return vs_version, url, sha_sum
    return None


async def get_package_details(static_root: Path) -> tuple[str, str, str]:
    """Fetch details about the latest version of the monaco editor.

    Parameters
    ----------
    static_root : Path
        The root directory for storing package metadata.

    Returns
    -------
    tuple[str, str, str]
        A tuple containing the latest version, tarball URL, and SHA-1 checksum.

    Raises
    ------
    DownloadError
        If fetching package details fails.
    """
    # let's merge loading and checking the file
    monaco_details = static_root / "monaco" / "monaco.json"
    monaco_details.parent.mkdir(parents=True, exist_ok=True)
    cached = check_cached_details(monaco_details)
    if cached:
        if not PINNED_VERSION or PINNED_VERSION == cached[0]:
            return cached
    try:
        async with httpx.AsyncClient(timeout=20) as client:
            response = await client.get(f"{REGISTRY_BASE_URL}/{PACKAGE_NAME}")
            if response.status_code != 200:
                raise DownloadError(
                    f"Failed to fetch package details: {response.status_code}"
                )

            package_data = response.json()
            target_version = package_data.get("dist-tags", {}).get("latest")
            if PINNED_VERSION and PINNED_VERSION in package_data.get(
                "versions"
            ):
                target_version = PINNED_VERSION
            version_info = package_data.get("versions", {}).get(
                target_version, "latest"
            )
            tarball_url = version_info.get("dist", {}).get("tarball")
            sha_sum = version_info.get("dist", {}).get("shasum")
            if not all((target_version, tarball_url, sha_sum)):
                raise DownloadError("Package details are incomplete.")

            # Cache the details with a 'last_check' timestamp
            monaco_details.write_text(
                json.dumps(
                    {
                        "version": target_version,
                        "url": tarball_url,
                        "sha_sum": sha_sum,
                        "last_check": datetime.now(timezone.utc).isoformat(),
                    },
                    indent=2,
                ),
                encoding="utf-8",
                newline="\n",
            )
            return target_version, tarball_url, sha_sum
    except BaseException as e:
        raise DownloadError(f"Failed to fetch package details: {e}") from e


def extract_monaco_tar(tar_data: bytes, monaco_path: Path) -> None:
    """Extract the monaco editor tarball.

    Parameters
    ----------
    tar_data : bytes
        The tarball data.
    monaco_path : Path
        The path to the monaco editor files.

    Raises
    ------
    FileNotFoundError
        If the extraction fails.
    """
    monaco_path.mkdir(parents=True, exist_ok=True)
    with tarfile.open(fileobj=io.BytesIO(tar_data)) as tar:
        if tar_has_filter_parameter():
            tar.extractall(monaco_path, filter="data")  # nosemgrep # nosec
        else:
            tar.extractall(monaco_path)  # nosemgrep # nosec
    monaco_editor_root = monaco_path / "package"
    vs_src = monaco_editor_root / "min" / "vs"
    if not vs_src.exists():
        raise FileNotFoundError("Failed to extract monaco editor files.")
    vs_dst = monaco_path / "vs"
    if vs_dst.exists():
        shutil.rmtree(vs_dst)
    shutil.move(vs_src, vs_dst)
    min_map_src = monaco_editor_root / "min-maps"
    if min_map_src.exists():
        min_map_dst = monaco_path / "min-maps"
        if min_map_dst.exists():
            shutil.rmtree(min_map_dst)
        shutil.move(min_map_src, min_map_dst)
    shutil.rmtree(monaco_editor_root)


async def download_monaco_editor(static_root: Path) -> None:
    """Download and extract the monaco editor files.

    Parameters
    ----------
    static_root : Path
        The root directory for the monaco editor files.

    Raises
    ------
    DownloadError
        If the download fails.

    ValueError
        If the SHA-1 checksum does not match the expected value.
    """
    details = await get_package_details(static_root)
    loader_js = static_root / "monaco" / "vs" / "loader.js"
    if loader_js.is_file():
        LOG.info("Monaco editor files are already present.")
        return
    _version, url, sha_sum = details
    try:
        async with httpx.AsyncClient(timeout=60) as client:
            response = await client.get(url)
            if response.status_code != 200:
                raise DownloadError(
                    f"Failed to download Monaco Editor: {response.status_code}"
                )

            tar_data = response.content
    except BaseException as e:
        raise DownloadError(f"Failed to download Monaco Editor: {e}") from e

    calculated_sha_sum = hashlib.sha1(
        tar_data, usedforsecurity=False
    ).hexdigest()
    if calculated_sha_sum != sha_sum:
        raise ValueError("SHA-1 checksum mismatch.")
    monaco_path = static_root / "monaco"
    try:
        extract_monaco_tar(tar_data, monaco_path)
    except BaseException as e:
        LOG.error("Failed to extract Monaco Editor files: %s", e)
        raise DownloadError("Failed to extract Monaco Editor files.") from e


async def download_swagger_assets(static_root: Path) -> None:
    """Download the Swagger UI assets.

    Parameters
    ----------
    static_root : Path
        The root directory for the Swagger UI files.

    Raises
    ------
    DownloadError
        If the download fails.
    """
    swagger_path = static_root / "swagger"
    swagger_path.mkdir(parents=True, exist_ok=True)
    try:
        async with httpx.AsyncClient(timeout=30) as client:
            for asset, subdir, url in SWAGGER_ASSETS:
                (swagger_path / subdir).mkdir(parents=True, exist_ok=True)
                asset_path = swagger_path / subdir / asset
                if asset_path.exists():
                    continue

                response = await client.get(url)
                if response.status_code != 200:
                    LOG.error(
                        "Failed to download %s: %s", asset, response.status_code
                    )
                    raise DownloadError(
                        f"Failed to download {asset}: {response.status_code}"
                    )

                asset_path.write_text(
                    response.text, encoding="utf-8", newline="\n"
                )
    except BaseException as e:
        LOG.error("Failed to download Swagger UI assets: %s", e)
        raise DownloadError("Failed to download Swagger UI assets.") from e


async def ensure_extra_static_files(static_root: Path) -> None:
    """Ensure all required extra static files are present.

    Parameters
    ----------
    static_root : Path
        The root directory for all static files.

    Raises
    ------
    DownloadError
        If any download fails.
    """
    frontend_dir = static_root / "frontend"
    frontend_dir.mkdir(parents=True, exist_ok=True)
    if not (frontend_dir / "index.html").exists():
        (frontend_dir / "index.html").write_text(
            "The frontend files are not found. Please build the frontend first."
        )
    try:
        await download_monaco_editor(static_root)
        LOG.info("Monaco editor files are up-to-date.")
    except DownloadError as e:
        LOG.error("Error ensuring monaco editor files: %s", e)
        raise DownloadError("Failed to download Monaco Editor.") from e

    try:
        await download_swagger_assets(static_root)
        LOG.info("Swagger UI assets are up-to-date.")
    except BaseException as e:
        LOG.error("Error ensuring Swagger UI assets: %s", e)
        raise DownloadError("Failed to download Swagger UI assets.") from e


def tar_has_filter_parameter() -> bool:  # pragma: no cover
    """Check if the tarfile.extractall method has the filter parameter.

    Returns
    -------
    bool
        True if the filter parameter is available, False otherwise.
    """
    # Changed in version 3.10.12: Added the filter parameter.
    # Changed in version 3.11.4: Added the filter parameter.
    # Changed in version 3.12: Added the filter parameter.
    py_version = version.parse(sys.version.split()[0])
    if py_version.minor >= 12:
        return True
    if py_version.minor == 11 and py_version.micro >= 4:
        return True
    if py_version.minor == 10 and py_version.micro >= 12:
        return True
    return False


if __name__ == "__main__":
    logging.basicConfig(stream=sys.stdout, level=logging.INFO)
    asyncio.run(ensure_extra_static_files(STATIC_ROOT))



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pylint: disable=broad-exception-caught
# pyright: reportUnknownArgumentType=false,reportArgumentType=false

"""Module for serializing results into a JSON-compatible format."""

import json
from dataclasses import asdict, is_dataclass
from typing import Any


def serialize_dict(data: dict[Any, Any]) -> list[dict[Any, Any]]:
    """Serialize a dictionary into a json-compatible format.

    Parameters
    ----------
    data : dict[Any, Any]
        The dictionary to serialize.

    Returns
    -------
    list[dict[Any, Any]]
        The serialized dictionary with JSON-compatible types.
    """
    serialized: dict[str, Any] = {}
    for key, value in data.items():
        # If value is a dataclass, recursively serialize it
        if is_dataclass_instance(value):  # Check if it's a dataclass
            serialized[key] = serialize_results(
                asdict(value)
            )  # Serialize dataclass object
        elif isinstance(value, dict):  # If it's a nested dict, serialize it
            serialized[key] = serialize_dict(value)[0]
        elif isinstance(value, list):  # If it’s a list, serialize it
            serialized[key] = serialize_list(value)
        else:
            # Primitive types (str, int, float, etc.) are already serializable
            serialized[key] = value
    return [serialized]  # Return as a list


def serialize_list(data: list[Any]) -> list[dict[Any, Any]]:
    """Serialize a list into a list of JSON-compatible dictionaries.

    Parameters
    ----------
    data : list[Any]
        The list to serialize.

    Returns
    -------
    list[dict[Any, Any]]
        The list of serialized items.
    """
    serialized: list[dict[Any, Any]] = []
    for item in data:
        if is_dataclass_instance(
            item
        ):  # If the item is a dataclass, serialize it
            serialized.extend(
                serialize_results(asdict(item))
            )  # Flatten the result
        elif isinstance(
            item, dict
        ):  # If the item is a dictionary, serialize it
            serialized.append(
                serialize_dict(item)[0]
            )  # Ensure it’s wrapped in a list
        elif isinstance(item, list):  # If the item is a list, serialize it
            # Flatten the result
            serialized.extend(serialize_list(item))
        elif isinstance(item, (str, int, float, bool)):
            # Primitive types (str, int, float, bool) are already serializable
            # noinspection PyTypeChecker
            serialized.append(item)  # type: ignore
        elif hasattr(item, "model_dump"):
            # If the item has a model_dump method
            # (e.g., Pydantic models), serialize it
            serialized.append(item.model_dump(mode="json"))
        else:
            serialized.append({"value": json.dumps(item, default=str)})
    return serialized


def serialize_results(results: Any) -> list[dict[Any, Any]]:
    """Make the results JSON serializable.

    Parameters
    ----------
    results : Any
        The results.

    Returns
    -------
    list[dict[Any, Any]]
        The json serializable results.
    """
    if isinstance(results, dict):
        return serialize_dict(results)
    if isinstance(results, list):
        return serialize_list(results)
    if is_dataclass_instance(results):
        return serialize_dict(asdict(results))
    return serialize_results([results])  # pragma: no cover


def is_dataclass_instance(obj: Any) -> bool:
    """Check if an object is an instance of a dataclass.

    Parameters
    ----------
    obj : Any
        The object.

    Returns
    -------
    bool
        True if the object is an instance of a dataclass, False otherwise.
    """
    return is_dataclass(obj) and not isinstance(obj, type)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Convert sync functions to async ones and vice-versa."""

import asyncio
import concurrent.futures
import contextvars
from collections.abc import Coroutine
from concurrent.futures import Executor
from functools import partial, wraps
from typing import Any, Callable, TypeVar

try:
    from typing import ParamSpec  # type: ignore[unused-ignore,assignment]
except ImportError:
    from typing_extensions import (  # type: ignore[unused-ignore,assignment]
        ParamSpec,
    )

P = ParamSpec("P")
R = TypeVar("R")


def sync_to_async(
    func: Callable[P, R], executor: Executor | None = None
) -> Callable[P, Coroutine[Any, Any, R]]:
    """Run a synchronous function in an asynchronous way.

    Parameters
    ----------
    func : Callable
        The synchronous function to run asynchronously.
    executor : Executor, Optional
        An optional executor to use

    Returns
    -------
    Callable
        The asynchronous function.
    """
    if asyncio.iscoroutinefunction(func):
        return func

    @wraps(func)
    async def runner(*args: P.args, **kwargs: P.kwargs) -> R:
        """Run the synchronous function in an executor.

        Parameters
        ----------
        *args : Any
            The positional arguments for the function.
        **kwargs : Any
            The keyword arguments for the function.

        Returns
        -------
        R
            The result of the synchronous function executed asynchronously.
        """
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            # No event loop running, execute synchronously
            return func(*args, **kwargs)

        partial_func = partial(func, *args, **kwargs)
        if executor is None:
            # noinspection PyTypeChecker
            return await asyncio.to_thread(partial_func)
        return await loop.run_in_executor(executor, partial_func)

    return runner


def async_to_sync(func: Callable[P, Coroutine[Any, Any, R]]) -> Callable[P, R]:
    """Run an asynchronous function in a synchronous way.

    Parameters
    ----------
    func : Callable
        The asynchronous function to run synchronously.

    Returns
    -------
    Callable
        The synchronous function.
    """

    @wraps(func)
    def run_sync(*args: P.args, **kwargs: P.kwargs) -> R:
        """Run the asynchronous function synchronously.

        Parameters
        ----------
        *args : Any
            The positional arguments for the function.
        **kwargs : Any
            The keyword arguments for the function.

        Returns
        -------
        R
            The result of the asynchronous function executed synchronously.
        """
        try:
            # Check if there's already an event loop running
            asyncio.get_running_loop()
        except RuntimeError:
            # No event loop running, we can create one
            return asyncio.run(func(*args, **kwargs))

        ctx = contextvars.copy_context()

        def _run_in_new_loop() -> R:
            new_loop = asyncio.new_event_loop()
            # pylint: disable=too-many-try-statements
            try:
                asyncio.set_event_loop(new_loop)
                # Create the coroutine *inside* the thread/loop context
                coro = func(*args, **kwargs)
                result = new_loop.run_until_complete(coro)
                # Gracefully shut down asynchronous generators (py>=3.6)
                new_loop.run_until_complete(new_loop.shutdown_asyncgens())
                return result
            finally:
                try:
                    asyncio.set_event_loop(None)
                finally:
                    new_loop.close()

        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            future = executor.submit(lambda: ctx.run(_run_in_new_loop))
            return future.result()

    return run_sync



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Subprocess engine implementation."""

from __future__ import annotations

import asyncio
import contextlib
import os
import signal
import subprocess
import sys
import time
from pathlib import Path
from typing import Any

from fastapi import WebSocket

from .base import Engine

_READ_CHUNK = 16_384  # 16 KB per read; safe and responsive
_MAX_PENDING_LINE_BYTES = 2_000_000  # safety: split very long lines


class SubprocessEngine(Engine):
    """Subprocess engine implementation."""

    file_path: Path
    root_dir: Path
    websocket: WebSocket

    proc: asyncio.subprocess.Process | None = None
    _monitor_tasks: list[asyncio.Task[Any]] | None = None
    _queue: asyncio.Queue[dict[str, Any]] | None = None
    _start_ts: float = 0.0
    _wait_task: asyncio.Task[Any] | None = None
    _did_end: bool = False

    async def start(self, start_msg: dict[str, Any] | None = None) -> None:
        """Start the run (after receiving the initial 'op=start' message).

        Parameters
        ----------
        start_msg : dict[str, Any]
            Optional start arguments to pass.
        """
        start_msg = start_msg or {}
        module = start_msg.get("module", "")
        args = list(start_msg.get("args", []))
        env = os.environ.copy()
        env.update(start_msg.get("env", {}))
        env.setdefault("PYTHONUNBUFFERED", "1")
        env.setdefault("PYTHONIOENCODING", "utf-8")
        module_args: list[str] = (
            ["-m", module] if module else [str(self.file_path)]
        )
        py = sys.executable
        venv = start_msg.get("venv")
        if venv:
            if sys.platform == "win32":  # pragma: no cover
                cmd = Path(venv) / "Scripts" / "python.exe"
            else:
                cmd = Path(venv) / "bin" / "python"
            if cmd.exists():
                py = str(cmd)

        cwd = (
            self.root_dir
            if not start_msg.get("cwd")
            else (self.root_dir / str(start_msg["cwd"]))
        )
        self._start_ts = time.time()
        creationflags = 0
        if sys.platform == "win32":
            creationflags = getattr(subprocess, "CREATE_NEW_PROCESS_GROUP", 0)
        self.proc = await asyncio.create_subprocess_exec(
            py,
            *module_args,
            *args,
            cwd=str(cwd),
            env=env,
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            start_new_session=sys.platform != "win32",
            creationflags=creationflags,
        )

        self._queue = asyncio.Queue(maxsize=10000)
        await self._queue.put(
            {
                "type": "run_status",
                "data": {
                    "state": "started",
                    "pid": self.proc.pid,
                    "cwd": str(cwd),
                },
            }
        )

        self._monitor_tasks = [
            asyncio.create_task(self._read_stream(self.proc.stdout, "stdout")),
            asyncio.create_task(self._read_stream(self.proc.stderr, "stderr")),
            asyncio.create_task(self._send_queued()),
        ]
        self._wait_task = asyncio.create_task(self._wait_and_finalize())

    def _interrupt(self) -> None:
        if not self.proc:
            return
        # pylint: disable=no-member
        try:
            if sys.platform == "win32":
                # Requires CREATE_NEW_PROCESS_GROUP at spawn time
                self.proc.send_signal(signal.CTRL_BREAK_EVENT)  # type: ignore
            else:
                try:
                    pgid = os.getpgid(self.proc.pid)
                    # cspell: disable-next-line
                    os.killpg(pgid, signal.SIGINT)
                except PermissionError:
                    self.proc.send_signal(signal.SIGINT)
        except ProcessLookupError:
            pass

    def _terminate(self) -> None:
        if not self.proc:
            return
        # pylint: disable=no-member,too-many-try-statements
        try:
            if sys.platform == "win32":
                self.proc.terminate()
                return
            getpgid = getattr(os, "getpgid", None)
            killpg = getattr(os, "killpg", None)

            if getpgid and killpg:
                try:
                    # pylint: disable=not-callable
                    pgid = getpgid(self.proc.pid)
                    killpg(pgid, signal.SIGTERM)
                    return
                except PermissionError:
                    # Fall back to terminating just the child
                    self.proc.terminate()
                    return
            # If no pgid/killpg, fall back
            self.proc.terminate()
        except ProcessLookupError:
            pass

    def _kill(self) -> None:
        if not self.proc:
            return
        # pylint: disable=no-member,too-many-try-statements
        try:
            if sys.platform == "win32":
                self.proc.kill()
                return
            getpgid = getattr(os, "getpgid", None)
            killpg = getattr(os, "killpg", None)
            sigkill = getattr(signal, "SIGKILL", signal.SIGTERM)

            if getpgid and killpg:
                try:
                    # pylint: disable=not-callable
                    pgid = getpgid(self.proc.pid)
                    killpg(pgid, sigkill)
                    return
                except PermissionError:
                    self.proc.kill()
        except ProcessLookupError:
            pass

    async def _handle_stdin(self, msg: dict[str, Any]) -> None:
        """Handle stdin request."""
        self.log.debug("Got message: %s", msg)
        if not self.proc:
            return
        # pylint: disable=too-many-try-statements,broad-exception-caught
        try:
            text = msg.get("text", "") or ""
            if not text.endswith("\n"):
                text += "\n"
            data = text.encode(errors="replace")
            if self.proc.stdin and not self.proc.stdin.is_closing():
                self.proc.stdin.write(data)
                await self.proc.stdin.drain()
                if self._queue:
                    await self._queue.put(
                        {
                            "type": "run_stdin_ack",
                            "data": {
                                "text": text,
                                "request_id": msg.get("request_id"),
                            },
                        }
                    )
            else:
                if self._queue:
                    await self._queue.put(
                        {
                            "type": "run_stdin_error",
                            "data": {
                                "message": msg,
                                "error": "No active process",
                            },
                        }
                    )
        except Exception as e:
            if self._queue:
                await self._queue.put(
                    {
                        "type": "run_stdin_error",
                        "data": {"message": msg, "error": str(e)},
                    }
                )

    # pylint: disable=too-complex
    async def handle_client(self, msg: dict[str, Any]) -> None:  # noqa: C901
        """Handle subsequent client control messages (stdin/interrupt/etc.).

        Parameters
        ----------
        msg : dict[str, Any]
            Message from the client.
        """
        self.log.debug("Handling client message: %s", msg)
        if not self.proc:
            return
        op = msg.get("op")

        if op == "stdin":
            await self._handle_stdin(msg)

        elif op == "stdin_eof":
            if self.proc.stdin and not self.proc.stdin.is_closing():
                with contextlib.suppress(Exception, NotImplementedError):
                    self.proc.stdin.write_eof()

        elif op == "interrupt":
            self._interrupt()

        elif op in ("terminate", "shutdown"):
            self._terminate()

        elif op == "kill":
            self._kill()

    async def shutdown(self) -> None:
        """Finalize the run and emit 'run_end' once via _finalize()."""
        # try to let the proc end quickly; if not, escalate a bit
        if self.proc and self.proc.returncode is None:
            try:
                await asyncio.wait_for(self.proc.wait(), timeout=5.0)
            except asyncio.TimeoutError:
                # gentle stop
                self._terminate()
                try:
                    await asyncio.wait_for(self.proc.wait(), timeout=3.0)
                except asyncio.TimeoutError:
                    # hard stop
                    self._kill()
                    with contextlib.suppress(Exception):
                        await self.proc.wait()

        # cancel the waiter so it doesn't race; then finalize exactly once
        if self._wait_task:
            self._wait_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._wait_task
            self._wait_task = None

        # compute rc and finalize (will cancel readers/sender and send run_end)
        rc = self.proc.returncode if self.proc else 0
        await self._finalize(rc)

    async def _wait_and_finalize(self) -> None:
        if not self.proc:
            return
        try:
            rc = await self.proc.wait()
        except asyncio.CancelledError:
            return
        await self._finalize(rc)

    async def _finalize(self, rc: int | None) -> None:
        if self._did_end:
            return
        self._did_end = True

        # stop readers/sender
        if self._monitor_tasks:
            for t in self._monitor_tasks:
                t.cancel()
            for t in self._monitor_tasks:
                with contextlib.suppress(asyncio.CancelledError, RuntimeError):
                    await t
            self._monitor_tasks = None

        elapsed = int((time.time() - self._start_ts) * 1000)
        await self._send_safe(
            {
                "type": "run_end",
                "data": {
                    "status": "ok" if (rc or 0) == 0 else "error",
                    "returnCode": rc or 0,
                    "elapsedMs": elapsed,
                },
            }
        )

    # pylint: disable=too-many-try-statements,too-complex
    async def _read_stream(  # noqa: C901
        self, stream: asyncio.StreamReader | None, label: str
    ) -> None:
        if not self.proc or not stream:
            self.log.warning(
                "No process or stream. no process: %s, no stream: %s",
                not self.proc,
                not stream,
            )
            return
        buf = bytearray()

        async def emit(line_bytes: bytes) -> None:
            # Decode defensively; your downstream can JSON-parse if it wants.
            text = line_bytes.decode("utf-8", errors="replace")
            if self._queue:
                await self._queue.put(
                    {
                        "type": f"run_{label}",  # "stdout" or "stderr"
                        "data": {"text": text},
                    }
                )

        try:
            while True:
                chunk = await stream.read(_READ_CHUNK)
                if not chunk:  # EOF
                    break
                buf.extend(chunk)

                # Emit every complete line we have
                start = 0
                while True:
                    nl = buf.find(b"\n", start)
                    if nl == -1:
                        # If we’ve buffered a massive no-newline line,
                        # spill a piece
                        if len(buf) > _MAX_PENDING_LINE_BYTES:
                            spill, buf = (
                                bytes(buf[:_READ_CHUNK]),
                                buf[_READ_CHUNK:],
                            )
                            await emit(spill)
                        break
                    line, start = bytes(buf[:nl]), nl + 1
                    await emit(line)
                # Keep only the remainder (after last newline)
                if start:
                    del buf[:start]

            # Flush any tail without a trailing newline
            if buf:
                await emit(bytes(buf))
        except asyncio.CancelledError as e:
            raise asyncio.CancelledError(str(e)) from e
        except Exception as e:  # pylint: disable=broad-exception-caught
            # Surface stream errors in the queue so callers can react
            if self._queue:
                await self._queue.put(
                    {
                        "type": f"{label}_error",
                        "data": f"{type(e).__name__}: {e}",
                    }
                )

    async def _send_queued(self) -> None:
        if not self._queue:
            return
        try:
            while True:
                try:
                    message = await asyncio.wait_for(
                        self._queue.get(), timeout=1.0
                    )
                except asyncio.TimeoutError:
                    # periodically check for process end to exit faster
                    # if queue empty
                    if (
                        self.proc
                        and self.proc.returncode is not None
                        and self._queue.empty()
                    ):
                        break
                    continue
                await self._send_safe(message)
        except asyncio.CancelledError:
            # flush a few messages quickly on cancellation
            for _ in range(50):
                if self._queue.empty():
                    break
                with contextlib.suppress(Exception):
                    await self._send_safe(self._queue.get_nowait())
            raise

    async def _send_safe(self, message: Any) -> None:
        with contextlib.suppress(Exception):
            if isinstance(message, (dict, list, tuple, set)):
                await self.websocket.send_json(message)
            else:
                await self.websocket.send_text(str(message))



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pylint: disable=broad-exception-caught,too-many-try-statements
# pyright: reportCallInDefaultInitializer=false

"""Terminal websocket route."""

import asyncio
import contextlib
import json
import os
from pathlib import Path

from fastapi import (
    APIRouter,
    Depends,
    HTTPException,
    WebSocket,
    WebSocketDisconnect,
)

from waldiez_studio.utils.terminal_session import get_session

from .common import get_root_directory

router = APIRouter()


def _safe_workdir(root: Path, rel: str | None) -> Path:
    base = root.resolve()
    target = (base / (rel or "")).resolve()
    if base == target or str(target).startswith(str(base) + os.sep):
        return target if target.exists() else base
    raise ValueError("cwd outside workspace")


# pylint: disable=too-complex,too-many-locals,
# pylint: disable=too-many-branches,too-many-statements
@router.websocket("/ws/terminal")
async def terminal_ws(  # noqa: C901
    ws: WebSocket,
    cwd: str | None = None,
    root_dir: Path = Depends(get_root_directory),
) -> None:
    """Terminal websocket endpoint.

    Parameters
    ----------
    ws : Websocket
        The websocket.
    cwd : str | None
        The working directory to use, defaults to None
    root_dir : Path
        The root directory dependency.

    Raises
    ------
    ValueError
        If the cwd is not inside the workspace.
    HTTPException
        If sth goes wrong.
    """
    await ws.accept()
    try:
        workdir = _safe_workdir(root_dir, cwd)
    except Exception as exc:
        # close with a clear reason
        with contextlib.suppress(Exception):
            await ws.close(code=1008, reason="Invalid cwd")
        raise HTTPException(400, "Invalid cwd") from exc

    session = get_session(workdir)

    stop = asyncio.Event()
    sent_end = False  # ensure we only notify end once

    async def _notify_end_once() -> None:
        nonlocal sent_end
        if sent_end:
            return
        sent_end = True
        with contextlib.suppress(Exception):
            # single, consistent end-of-session event
            await ws.send_json({"type": "session_end"})

    async def _reader() -> None:
        """Pump PTY -> WS."""
        try:  # pragma: no branch
            while not stop.is_set():
                data = await session.read(4096)
                if not data:
                    # allow other tasks to run & check liveness
                    await asyncio.sleep(0.01)
                    if not session.is_alive():
                        await _notify_end_once()
                        break
                    continue
                with contextlib.suppress(Exception):
                    await ws.send_json(
                        {"type": "data", "data": data.decode("utf-8", "ignore")}
                    )
        except asyncio.CancelledError:
            # normal on ws/tab close or shutdown
            pass
        except Exception:  # pragma: no cover
            # swallow—socket may be closing
            pass
        finally:
            stop.set()

    read_task = asyncio.create_task(_reader())

    # process an initial message (e.g., resize from client)
    try:
        with contextlib.suppress(Exception):
            msg = await asyncio.wait_for(ws.receive_text(), timeout=1.0)
            payload = json.loads(msg)
            if payload.get("op") == "resize":
                session.resize(
                    int(payload.get("rows") or 24),
                    int(payload.get("cols") or 80),
                )
    except Exception:  # pragma: no cover
        pass

    try:
        while not stop.is_set():
            try:
                raw = await ws.receive_text()
            except WebSocketDisconnect:
                break
            except Exception:  # pragma: no cover
                break

            try:
                payload = json.loads(raw)
            except Exception:  # nosemgrep # nosec # pragma: no cover
                continue

            op = payload.get("op")
            if op == "stdin":
                data = payload.get("data") or ""
                session.write(data.encode("utf-8", "ignore"))
            elif op == "resize":
                session.resize(
                    int(payload.get("rows") or 24),
                    int(payload.get("cols") or 80),
                )
            elif op == "interrupt":
                session.interrupt()
            elif op in ("terminate", "kill"):
                session.terminate()
                await _notify_end_once()
                break
    finally:
        stop.set()
        # cancel reader and ignore cancellation/context shutdown
        if not read_task.done():
            with contextlib.suppress(Exception):
                read_task.cancel()
            with contextlib.suppress(asyncio.CancelledError, Exception):
                await read_task

        await _notify_end_once()
        session.close()
        with contextlib.suppress(Exception):
            await ws.close()



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pyright: reportUnknownVariableType=false
"""Waldiez engine implementation."""

from __future__ import annotations

import asyncio
import contextlib
import json
from pathlib import Path
from typing import Any

from waldiez import WaldiezExporter

from .base import Engine
from .subprocess_engine import SubprocessEngine


class WaldiezEngine(Engine):
    """Waldiez engine implementation."""

    _task: asyncio.Task[Any] | None = None
    _delegate: SubprocessEngine | None = None

    async def start(self, start_msg: dict[str, Any] | None = None) -> None:
        """Start the run (after receiving the initial 'op=start' message).

        Parameters
        ----------
        start_msg : dict[str, Any]
            Optional start arguments to pass.
        """
        start_msg = start_msg or {}
        await self._send_safe(
            {
                "type": "compile_start",
                "data": {
                    "source": str(self.file_path.relative_to(self.root_dir))
                },
            }
        )
        try:
            py_path = await self._compile_to_py(self.file_path)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            await self._send_safe(
                {"type": "compile_error", "data": {"message": f"{exc}"}}
            )
            # Surface an early end for the UI; no delegate run
            await self._send_safe(
                {
                    "type": "run_end",
                    "data": {
                        "status": "error",
                        "returnCode": -1,
                        "elapsedMs": 0,
                    },
                }
            )
            return

        await self._send_safe(
            {
                "type": "compile_end",
                "data": {"py": str(py_path.relative_to(self.root_dir))},
            }
        )

        # Delegate to the subprocess engine
        self._delegate = SubprocessEngine(
            file_path=py_path, root_dir=self.root_dir, websocket=self.websocket
        )
        args: dict[str, Any] = {
            "module": "waldiez",  # python -m waldiez
            "args": [
                "run",
                "--file",
                self.file_path,
                "--output",
                py_path,
                "--force",
                "--structured",
            ],
        }
        cmd_args = start_msg.get("args", [])
        if cmd_args and isinstance(cmd_args, list):
            for cmd_arg in cmd_args:
                if isinstance(cmd_arg, str):
                    args["args"].append(cmd_arg)
        await self._delegate.start(args)

    async def handle_client(self, msg: dict[str, Any]) -> None:
        """Handle subsequent client control messages (stdin/interrupt/etc.).

        Parameters
        ----------
        msg : dict[str, Any]
            Message from the client.
        """
        if not self._delegate:
            self.log.error("No delegate for the process.")
            return
        the_msg = msg
        if the_msg.get("op") in (
            "waldiez_respond",
            "waldiez_control",
        ) and the_msg.get("payload", {}):
            await self._delegate.handle_client(
                {"op": "stdin", "text": json.dumps(the_msg.get("payload"))}
            )
        else:
            await self._delegate.handle_client(msg)

    async def shutdown(self) -> None:
        """Finalize the run and emit 'run_end' if appropriate."""
        if self._task:
            self._task.cancel()
            with contextlib.suppress(asyncio.CancelledError, RuntimeError):
                await self._task
        self._task = None
        if self._delegate:
            await self._delegate.shutdown()

    @staticmethod
    async def _compile_to_py(src: Path) -> Path:
        exporter = WaldiezExporter.load(src)
        output = src.with_suffix(".py")
        exporter.export(output, force=True)
        return output

    async def _send_safe(self, message: dict[str, Any]) -> None:
        try:
            await self.websocket.send_json(message)
        except Exception:  # pylint: disable=broad-exception-caught
            # ignore socket issues; upstream will close
            pass



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=broad-exception-caught,too-complex,import-error
# flake8: noqa: C901
"""Terminal session."""

import sys
from pathlib import Path

from .base import BaseSession


def get_session(workdir: Path) -> BaseSession:  # pragma: no cover
    """Get a session based on the host's os.

    Parameters
    ----------
    workdir : Path
        The path to start the session with.

    Returns
    -------
    BaseSession
        A session based on the host's os.
    """
    # pylint: disable=import-outside-toplevel
    if sys.platform == "win32":
        from .windows_session import WindowsSession

        return WindowsSession(workdir)
    from .unix_session import UnixSession

    return UnixSession(workdir)


__all__ = [
    "BaseSession",
    "get_session",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pyright: reportUnnecessaryIsInstance=false, reportUnreachable=false

"""Config related helper functions.

To get configuration values from
environment variables or command line arguments.
"""

import os
import sys

ENV_PREFIX = "WALDIEZ_STUDIO_"


def get_trusted_hosts(domain_name: str, host: str) -> list[str]:
    """Get the trusted hosts.

    Parameters
    ----------
    domain_name : str
        The domain name

    host : str
        The host to listen on

    Returns
    -------
    list[str]
        The trusted hosts
    """
    from_env = os.environ.get(f"{ENV_PREFIX}TRUSTED_HOSTS", "")
    trusted_hosts: list[str] = from_env.split(",") if from_env else []
    if not isinstance(trusted_hosts, list):  # pragma: no cover
        # noinspection PyTypeChecker
        trusted_hosts = [trusted_hosts]
    if (
        not trusted_hosts or domain_name not in trusted_hosts
    ):  # pragma: no cover
        trusted_hosts.append(domain_name)
    if host not in trusted_hosts and host not in ["localhost", "0.0.0.0"]:
        trusted_hosts.append(host)
    if "--trusted-hosts" in sys.argv:  # pragma: no branch
        trusted_host_index = sys.argv.index("--trusted-hosts") + 1
        if trusted_host_index < len(sys.argv):  # pragma: no branch
            trusted_hosts_arg = sys.argv[trusted_host_index]
            if trusted_hosts_arg:  # pragma: no branch
                trusted_hosts_split = trusted_hosts_arg.split(",")
                for trusted_host in trusted_hosts_split:
                    if (
                        trusted_host and trusted_host not in trusted_hosts
                    ):  # pragma: no branch
                        trusted_hosts.append(trusted_host)
    return trusted_hosts


def get_trusted_origins(
    domain_name: str, port: int, force_ssl: bool, host: str
) -> list[str]:
    """Get the trusted origins.

    Parameters
    ----------
    domain_name : str
        The domain name
    port : int
        The port
    force_ssl : bool
        Whether to force SSL
    host : str
        The host to listen on

    Returns
    -------
    List[str]
        The trusted origins
    """
    from_env = os.environ.get(f"{ENV_PREFIX}TRUSTED_ORIGINS", "")
    trusted_origins = from_env.split(",") if from_env else []

    default_trusted_origins = [f"https://{domain_name}"]
    if host != domain_name:
        default_trusted_origins.append(f"https://{host}")
    if not force_ssl:
        # noinspection HttpUrlsUsage
        default_trusted_origins.extend(
            [
                f"http://{domain_name}",
                f"http://{domain_name}:{port}",
                f"http://{host}",
                f"http://{host}:{port}",
            ]
        )

    trusted_origins.extend(
        origin
        for origin in default_trusted_origins
        if origin not in trusted_origins
    )

    if "--trusted-origins" in sys.argv:  # pragma: no branch
        trusted_origin_index = sys.argv.index("--trusted-origins") + 1
        if trusted_origin_index < len(sys.argv):  # pragma: no branch
            trusted_origin = sys.argv[trusted_origin_index]
            if (
                trusted_origin and trusted_origin not in trusted_origins
            ):  # pragma: no branch
                trusted_origins.append(trusted_origin)

    return trusted_origins


def get_default_domain_name() -> str:
    """Get the default domain name.

    Returns
    -------
    str
        The default domain name
    """
    if "--domain-name" in sys.argv:  # pragma: no branch
        domain_name_index = sys.argv.index("--domain-name") + 1
        if domain_name_index < len(sys.argv):  # pragma: no branch
            domain_name = sys.argv[domain_name_index]
            if domain_name:  # pragma: no branch
                os.environ[f"{ENV_PREFIX}DOMAIN_NAME"] = domain_name
                return domain_name
    return os.environ.get(f"{ENV_PREFIX}DOMAIN_NAME", "localhost")


def get_default_host() -> str:
    """Get the default host.

    Returns
    -------
    str
        The default host
    """
    if "--host" in sys.argv:  # pragma: no branch
        host_index = sys.argv.index("--host") + 1
        if host_index < len(sys.argv):  # pragma: no branch
            host = sys.argv[host_index]
            if host:  # pragma: no branch
                os.environ[f"{ENV_PREFIX}HOST"] = host
                return host
    return os.environ.get(f"{ENV_PREFIX}HOST", "localhost")


def get_default_port() -> int:
    """Get the default port.

    Returns
    -------
    int
        The default port
    """
    if "--port" in sys.argv:  # pragma: no branch
        port_index = sys.argv.index("--port") + 1
        if port_index < len(sys.argv):  # pragma: no branch
            try:
                port = int(sys.argv[port_index])
                os.environ[f"{ENV_PREFIX}PORT"] = str(port)
                return port
            except ValueError:
                pass
    try:
        return int(os.environ.get(f"{ENV_PREFIX}PORT", "8000"))
    except ValueError:
        return 8000



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=too-many-try-statements,broad-exception-caught
"""Task runner to delegate the task to the appropriate engine."""

import asyncio
import contextlib
import json
import logging
import traceback
from typing import Any

from fastapi import WebSocket
from starlette.websockets import WebSocketDisconnect

from waldiez_studio.engines import Engine

LOG = logging.getLogger(__name__)

MAX_ACTIVE_TASKS = 10
_active_tasks = asyncio.Semaphore(MAX_ACTIVE_TASKS)


class TaskRunner:
    """Task runner to delegate the task to the appropriate engine."""

    def __init__(
        self, task_id: str, websocket: WebSocket, engine: Engine
    ) -> None:
        self.task_id = task_id
        self.websocket = websocket
        self.engine = engine

    # pylint: disable=too-complex
    async def listen(self) -> None:  # noqa: C901
        """Start listening.

        Raises
        ------
        asyncio.CancelledError
            If the task is cancelled.
        """
        async with _active_tasks:  # limit concurrent runs
            try:
                # First message must be op=start
                try:
                    raw = await self.websocket.receive_text()
                    start_msg = json.loads(raw)
                except Exception:
                    await self.send(
                        {
                            "type": "error",
                            "data": {"message": "invalid first message"},
                        }
                    )
                    return

                if start_msg.get("op") != "start":
                    await self.send(
                        {
                            "type": "error",
                            "data": {
                                "message": "first message must be op=start"
                            },
                        }
                    )
                    return

                await self.engine.start(start_msg)

                while True:
                    try:
                        raw = await self.websocket.receive_text()
                        msg = json.loads(raw)
                    except WebSocketDisconnect:
                        LOG.debug("WS disconnected")
                        break
                    except Exception:
                        await self.send(
                            {
                                "type": "error",
                                "data": {"message": "invalid message"},
                            }
                        )
                        continue

                    await self.engine.handle_client(msg)
                    # if msg.get("op") == "shutdown":
                    #     break

            except WebSocketDisconnect:  # pragma: no cover
                LOG.info("WS disconnected early")
            except asyncio.CancelledError:
                LOG.debug("TaskRunner cancelled")
                raise
            except Exception as exc:
                traceback.print_exc()
                LOG.warning("TaskRunner error: %s", exc)
            finally:
                with contextlib.suppress(Exception):
                    await self.engine.shutdown()

    async def send(self, payload: dict[str, Any]) -> None:
        """Safe send a message.

        Parameters
        ----------
        payload : dict[str, Any]
            The message to send.
        """
        with contextlib.suppress(Exception):
            await self.websocket.send_json(payload)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pyright: reportCallInDefaultInitializer=false

"""Common utilities for the api routes."""

import logging
import re
from pathlib import Path
from urllib.parse import unquote

from fastapi import Depends, HTTPException, Query

from waldiez_studio.utils.paths import get_root_dir

LOG = logging.getLogger(__name__)

ALLOWED_EXTENSIONS: dict[str, str] = {
    ".waldiez": "application/x-waldiez",
    ".bash": "text/x-shellscript",
    ".sh": "text/x-shellscript",
    ".zsh": "text/x-shellscript",
    ".ps1": "text/x-shellscript",
    ".bat": "text/x-shellscript",
    ".txt": "text/plain",
    ".py": "text/x-python",
    ".js": "application/javascript",
    ".ts": "application/typescript",
    ".md": "text/markdown",
    ".mmd": "text/x-mermaid",
    ".csv": "text/csv",
    ".rst": "text/x-rst",
    ".css": "text/css",
    ".html": "text/html",
    ".xml": "application/xml",
    ".json": "application/json",
    ".yaml": "application/x-yaml",
    ".yml": "application/x-yaml",
    ".toml": "application/toml",
    ".ini": "text/ini",
    ".ipynb": "application/x-ipynb+json",
    ".png": "image/png",
    ".jpg": "image/jpeg",
    ".jpeg": "image/jpeg",
    ".gif": "image/gif",
    ".webp": "image/webp",
    ".svg": "image/svg+xml",
    ".mp4": "video/mp4",
    ".webm": "video/webm",
    ".ogg": "video/ogg",
    ".mp3": "audio/mpeg",
    ".wav": "audio/wav",
    ".oga": "audio/ogg",
}
TEXTUAL_EXTS = {
    ".sh",
    ".bash",
    ".bsh",
    ".zsh",
    ".ps1",
    ".bat",
    ".txt",
    ".py",
    ".js",
    ".csv",
    ".ts",
    ".md",
    ".mmd",
    ".rst",
    ".css",
    ".html",
    ".xml",
    ".json",
    ".yaml",
    ".yml",
    ".toml",
    ".ini",
    ".ipynb",
    ".waldiez",
}


def get_root_directory() -> Path:
    """Get the root directory of the workspace.

    Returns
    -------
    Path
        The root directory of the workspace.
    """
    return get_root_dir()


def _resolve_path(path: Path, strict: bool = False) -> Path:
    """Resolve a path and ensure it exists.

    Parameters
    ----------
    path : Path
        The path to resolve.
    strict : bool, optional
        Whether to strictly enforce the path, by default False.

    Returns
    -------
    Path
        The resolved path.

    Raises
    ------
    FileNotFoundError
        If the path does not exist.
    """
    try:
        resolved_path = path.resolve(strict=strict)
    except FileNotFoundError as error:
        raise FileNotFoundError("Error: Path not resolved") from error
    except BaseException as error:
        raise ValueError("Error: Path not resolved") from error
    return resolved_path


def sanitize_path(root_dir: Path, path: str, strict: bool = False) -> Path:
    """Sanitize a path and ensure it is within the root directory.

    Parameters
    ----------
    root_dir : Path
        The root directory to check against.
    path : str
        The path to sanitize.
    strict : bool, optional
        Whether to strictly enforce the path, by default False.

    Returns
    -------
    Path
        The sanitized path.

    Raises
    ------
    ValueError
        If the path is invalid or traverses outside the root directory.
    BaseException
        If the path cannot be resolved.
    """
    if path in ("/", ""):
        return root_dir
    try:
        unquoted_path = unquote(path, errors="strict")
        if not unquoted_path.strip():
            return root_dir
    except BaseException as error:
        raise ValueError("Error: Invalid path") from error
    if unquoted_path in ("/", ""):
        return root_dir
    # Reject paths with invalid characters
    if re.search(r"[<>:\"|?*]", unquoted_path):
        raise ValueError("Error: Invalid path")
    safe_path = root_dir / Path(unquoted_path.strip("/"))
    try:
        resolved_path = _resolve_path(safe_path, strict=strict)
    except BaseException as error:
        raise error
    if not resolved_path.is_relative_to(root_dir):
        raise ValueError("Error: Invalid path: Path is outside root directory")
    return resolved_path


def check_path(
    path: str,
    root_dir: Path,
    path_type: str = "Path",
    must_exist: bool = True,
    must_not_exist: bool = False,
    must_be_dir: bool = False,
    must_be_file: bool = False,
    must_have_extension: str | tuple[str] | set[str] | None = None,
) -> Path:
    """Reusable dependency for sanitizing paths.

    Parameters
    ----------
    path : str
        The path to sanitize.
    root_dir : Path
        The root directory to use.
    path_type : str, optional
        The type of path, by default "Path" ("Path", "File" or "Directory").
    must_exist : bool, optional
        Whether the path must exist, by default True.
    must_not_exist : bool, optional
        Whether the path must not exist, by default False.
    must_be_dir : bool, optional
        Whether the path must be a directory, by default False.
    must_be_file : bool, optional
        Whether the path must be a file, by default False.
    must_have_extension : str | tuple[str] | set[str] | None
        The extension(s) the path must have, by default None.

    Returns
    -------
    Path
        The sanitized path.

    Raises
    ------
    HTTPException
        If the path is invalid.
    """
    thing = "File" if must_be_file else "Directory" if must_be_dir else "Path"
    try:
        the_path = sanitize_path(root_dir, path, must_exist)
    except FileNotFoundError as error:
        LOG.warning("Invalid path: %s, %s, %s", path, root_dir, error)
        raise HTTPException(
            status_code=404, detail=f"Error: {thing} not found"
        ) from error
    except ValueError as error:
        LOG.warning("Invalid path: %s, %s, %s", path, root_dir, error)
        raise HTTPException(
            status_code=400, detail=f"Error: Invalid {path_type.lower()}"
        ) from error
    thing = (
        _guess_path_type(must_be_dir, must_be_file, the_path)
        if path_type == "Path"
        else path_type
    )
    if must_exist and not the_path.exists():
        raise HTTPException(status_code=404, detail=f"Error: {thing} not found")
    if must_not_exist and the_path.exists():
        raise HTTPException(
            status_code=400, detail=f"Error: {thing} already exists"
        )
    if must_be_dir and not the_path.is_dir():
        raise HTTPException(status_code=400, detail="Error: Not a directory")
    if must_be_file and not the_path.is_file():
        raise HTTPException(status_code=400, detail="Error: Not a file")
    if must_have_extension is not None:
        extensions = (
            {must_have_extension}
            if isinstance(must_have_extension, str)
            else must_have_extension
        )
        if the_path.suffix not in extensions:
            raise HTTPException(
                status_code=400, detail="Error: Invalid file type"
            )
    return the_path.resolve(must_exist)


def _guess_path_type(
    should_be_dir: bool, should_be_file: bool, path: Path
) -> str:
    """Get the path type for error messages.

    Parameters
    ----------
    should_be_dir : bool
        Whether the path is expected to be a directory.
    should_be_file : bool
        Whether the path is expected to be a file.
    path : Path
        The path to check.

    Returns
    -------
    str
        The path type.
    """
    if should_be_dir:
        return "Directory"
    if should_be_file:
        return "File"
    if path.is_dir():
        return "Directory"
    if path.is_file():
        return "File"
    return "Path"


def get_new_folder_name(root_dir: Path, folder_name: str) -> str:
    """Get a new folder name that does not already exist.

    Parameters
    ----------
    root_dir : Path
        The root directory to check for existing folders.
    folder_name : str
        The original folder name.

    Returns
    -------
    str
        The new folder name.
    """
    # add (n) to the folder name if it already exists
    n = 0
    new_folder_name = folder_name
    while (root_dir / new_folder_name).exists():
        n += 1
        new_folder_name = f"{folder_name} ({n})"
    return new_folder_name


def get_new_file_name(root_dir: Path, file_name: str) -> str:
    """Get a new file name that does not already exist.

    Parameters
    ----------
    root_dir : Path
        The root directory to check for existing files.
    file_name : str
        The original file name.

    Returns
    -------
    str
        The new file name.
    """
    # add (n) to the file name if it already exists
    n = 0
    new_file_name = file_name
    while (root_dir / new_file_name).exists():
        n += 1
        parts = file_name.rsplit(".", maxsplit=1)
        if len(parts) == 2:
            new_file_name = f"{parts[0]} ({n}).{parts[1]}"
        else:
            new_file_name = f"{file_name} ({n})"
    return new_file_name


def check_flow_path(
    path: str = Query(..., description="The path to the flow."),
    root_dir: Path = Depends(get_root_directory),
) -> Path:
    """Validate the flow path.

    Parameters
    ----------
    path : str
        The path to the flow.
    root_dir : Path
        The root directory of the workspace.

    Returns
    -------
    Path
        The validated flow path relative to the root directory.
    Raises
    ------
    HTTPException
        If the path is invalid, the file does not exist,
        the file is not a file, or the file type is invalid.
    """
    return check_path(
        path,
        root_dir,
        path_type="File",
        must_exist=True,
        must_be_file=True,
        must_be_dir=False,
        must_have_extension=".waldiez",
    )



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
# pylint: disable=broad-exception-caught,too-complex,import-error,no-member
# flake8: noqa: C901
# pyright: reportUnknownMemberType=false,reportAttributeAccessIssue=false
# pyright: reportPossiblyUnboundVariable=false,reportUnknownVariableType=false
# pyright: reportUnknownArgumentType=false
"""Unix terminal session implementation."""

import asyncio
import contextlib
import fcntl
import os
import pty
import signal
import struct
import termios
from pathlib import Path

from .base import BaseSession


class UnixSession(BaseSession):
    """Unix session implementation."""

    ALLOWED_SHELLS = {"/bin/bash", "/bin/zsh", "/bin/sh", "/usr/bin/fish"}

    def __init__(self, workdir: Path):
        self.pid: int
        self.master_fd: int
        try:
            self.pid, self.master_fd = pty.fork()
        except OSError as e:
            raise RuntimeError(f"Failed to create PTY: {e}") from e
        if self.pid == 0:  # child  # pragma: no cover
            os.chdir(str(workdir))
            shell = os.environ.get("SHELL", "/bin/bash")
            if shell not in UnixSession.ALLOWED_SHELLS:
                shell = "/bin/bash"
            os.execvp(shell, [shell, "-l"])  # nosemgrep # nosec

        # parent
        flags = fcntl.fcntl(self.master_fd, fcntl.F_GETFL)
        # cspell: disable-next-line
        fcntl.fcntl(self.master_fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)
        self.resize(24, 80)

    def _set_win_size(self, rows: int, cols: int) -> None:
        fcntl.ioctl(
            self.master_fd,
            termios.TIOCSWINSZ,
            struct.pack("HHHH", rows, cols, 0, 0),
        )

    async def read(self, n: int = 4096) -> bytes:
        """Read.

        Parameters
        ----------
        n : int
            The number of rows to read.

        Returns
        -------
        bytes
            The read data.
        """
        # non-blocking read via thread offload
        try:
            return await asyncio.to_thread(os.read, self.master_fd, n)
        except BlockingIOError:
            return b""
        except OSError:
            return b""

    def write(self, data: bytes) -> None:
        """Write.

        Parameters
        ----------
        data : bytes
            The data to pass.
        """
        with contextlib.suppress(OSError):
            os.write(self.master_fd, data)

    def resize(self, rows: int, cols: int) -> None:
        """Resize the terminal.

        Parameters
        ----------
        rows : int
            The new number of rows.
        cols : int
            The new number of columns.
        """
        with contextlib.suppress(Exception):
            self._set_win_size(rows, cols)
            # cspell: disable-next-line
            os.kill(self.pid, signal.SIGWINCH)

    def interrupt(self) -> None:
        """Interrupt the session."""
        with contextlib.suppress(ProcessLookupError, PermissionError):
            # send to child only (group may require privileges)
            os.kill(self.pid, signal.SIGINT)

    def terminate(self) -> None:
        """Terminate the session."""
        with contextlib.suppress(ProcessLookupError):
            os.kill(self.pid, signal.SIGTERM)

    def is_alive(self) -> bool:
        """Check if the session is alive.

        Returns
        -------
        bool
            True if the session is alive, false otherwise.
        """
        with contextlib.suppress(ChildProcessError):
            # cspell: disable-next-line
            res = os.waitpid(self.pid, os.WNOHANG)
            return res == (0, 0)
        return False

    def close(self) -> None:
        """Close the session."""
        with contextlib.suppress(Exception):
            os.close(self.master_fd)
        with contextlib.suppress(ProcessLookupError):
            os.kill(self.pid, signal.SIGHUP)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pylint: disable=broad-exception-caught,import-error
# pyright: reportUnknownMemberType=false,reportAttributeAccessIssue=false
# pyright: reportPossiblyUnboundVariable=false,reportUnknownVariableType=false
# pyright: reportUnknownArgumentType=false,reportUnnecessaryIsInstance=false
# pyright: reportUnknownParameterType=false,reportMissingImports=false
# pyright: reportMissingTypeStubs=false

"""Windows terminal session implementation."""

import asyncio
import contextlib
from pathlib import Path
from shutil import which

from .base import BaseSession

try:
    import winpty  # type: ignore

    _HAVE_PYWINPTY = True
except (ImportError, ModuleNotFoundError):
    _HAVE_PYWINPTY = False  # pyright:ignore[reportConstantRedefinition]


def _preferred_windows_shell_argv() -> list[str]:
    """Prefer pwsh, then Windows PowerShell, then cmd."""
    candidates: tuple[tuple[str, list[str]], ...] = (
        ("pwsh.exe", ["-NoLogo", "-NoProfile"]),
        ("pwsh", ["-NoLogo", "-NoProfile"]),
        ("powershell.exe", ["-NoLogo", "-NoProfile"]),
        ("powershell", ["-NoLogo", "-NoProfile"]),
        ("cmd.exe", []),
        ("cmd", []),
    )
    for name, args in candidates:
        exe = which(name)
        if exe:
            return [exe, *args]
    # if which() finds nothing; let CreateProcess resolve it
    return ["powershell.exe", "-NoLogo", "-NoProfile"]


def _spawn_winpty(argv: list[str], cwd: str) -> winpty.PtyProcess:
    """
    Spawn using argv list if supported; otherwise join into a string.
    Avoid quoting the exe path.
    """
    try:
        return winpty.PtyProcess.spawn(argv, cwd=cwd)
    except TypeError:
        cmd = " ".join(argv)
        return winpty.PtyProcess.spawn(cmd, cwd=cwd)


class WindowsSession(BaseSession):
    """Windows session implementation."""

    def __init__(self, workdir: Path):
        if not _HAVE_PYWINPTY:
            msg = (
                "pywinpty is required for terminal on Windows. "
                "Install with: pip install pywinpty"
            )
            raise RuntimeError(msg)

        cwd = str(workdir)
        argv = _preferred_windows_shell_argv()

        # Optional: early sanity check & clearer error
        exe = Path(argv[0])
        if (exe.drive or exe.anchor) and not exe.is_file():
            # Don't include extra quotes here — keep the raw path
            raise FileNotFoundError(
                f"The shell executable was not found: {exe}"
            )

        self.pty = _spawn_winpty(argv, cwd=cwd)

    async def read(self, n: int = 4096) -> bytes:
        """Read.

        Parameters
        ----------
        n : int
            The number of rows to read.

        Returns
        -------
        bytes
            The read data.
        """
        try:
            data = await asyncio.to_thread(self.pty.read, n)
            return (
                data.encode("utf-8", "ignore")
                if isinstance(data, str)
                else data
            )
        except Exception:
            return b""

    def write(self, data: bytes) -> None:
        """Write.

        Parameters
        ----------
        data : bytes
            The data to pass.
        """
        try:
            self.pty.write(data.decode("utf-8", "ignore"))
        except Exception:
            pass

    def resize(self, rows: int, cols: int) -> None:
        """Resize the terminal.

        Parameters
        ----------
        rows : int
            The new number of rows.
        cols : int
            The new number of columns.
        """
        with contextlib.suppress(Exception):
            # cspell: disable-next-line
            self.pty.setwinsize(rows=rows, cols=cols)

    def interrupt(self) -> None:
        """Interrupt the session."""
        # Send Ctrl-C (0x03)
        with contextlib.suppress(Exception):
            self.pty.write("\x03")

    def terminate(self) -> None:
        """Terminate the session."""
        with contextlib.suppress(Exception):
            self.pty.terminate()

    def is_alive(self) -> bool:
        """Check if the session is alive.

        Returns
        -------
        bool
            True if the session is alive, false otherwise.
        """
        try:
            # cspell: disable-next-line
            return self.pty.isalive()
        except Exception:
            return False

    def close(self) -> None:
        """Close the session."""
        with contextlib.suppress(Exception):
            self.pty.close()



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Abstract execution engine interface."""

import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

from fastapi import WebSocket


class Engine(ABC):
    """Abstract execution engine interface."""

    file_path: Path
    root_dir: Path
    websocket: WebSocket

    def __init__(
        self, *, file_path: Path, root_dir: Path, websocket: WebSocket
    ) -> None:
        self.file_path = file_path
        self.root_dir = root_dir
        self.websocket = websocket
        self.log = logging.getLogger(self.__class__.__name__)

    @abstractmethod
    async def start(self, start_msg: dict[str, Any] | None = None) -> None:
        """Start the run (after receiving the initial 'op=start' message).

        Parameters
        ----------
        start_msg : dict[str, Any]
            Optional start arguments to pass.
        """

    @abstractmethod
    async def handle_client(self, msg: dict[str, Any]) -> None:
        """Handle subsequent client control messages (stdin/interrupt/etc.).

        Parameters
        ----------
        msg : dict[str, Any]
            Message from the client.
        """

    @abstractmethod
    async def shutdown(self) -> None:
        """Finalize the run and emit 'run_end' if appropriate."""



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pyright: reportUnknownMemberType=false

"""Kernel singleton instance."""

import asyncio
import time

from jupyter_client.manager import AsyncKernelManager


class KernelManager:
    """Kernel singleton instance."""

    _km: AsyncKernelManager | None = None
    _lock = asyncio.Lock()
    _last_used = 0.0
    IDLE_TTL_SEC = 15 * 60  # shutdown if idle for 15 min

    @classmethod
    async def get(cls) -> AsyncKernelManager:
        """Get the kernel manager.

        Returns
        -------
        AsyncKernelManager
            The kernel manager.
        """
        async with cls._lock:
            if cls._km is None:
                km = AsyncKernelManager(kernel_name="python3")
                await km.start_kernel()
                cls._km = km
            cls._last_used = time.time()
        return cls._km or AsyncKernelManager(kernel_name="python3")

    @classmethod
    async def interrupt(cls) -> None:
        """Interrupt the kernel."""
        if cls._km:  # pragma: no branch
            await cls._km.interrupt_kernel()

    @classmethod
    async def restart(cls) -> None:
        """Restart the kernel."""
        if cls._km:  # pragma: no branch
            await cls._km.restart_kernel(now=True)

    @classmethod
    async def shutdown_kernel(cls, now: bool = False) -> None:
        """Explicitly shutdown the kernel.

        Parameters
        ----------
        now : bool
            Shutdown now instead of waiting for idle.
        """
        async with cls._lock:
            if cls._km:
                await cls._km.shutdown_kernel(now=now)
                cls._km = None

    @classmethod
    async def maybe_gc(cls) -> None:
        """Call periodically (e.g., in a background task)."""
        if cls._km and time.time() - cls._last_used > cls.IDLE_TTL_SEC:
            await cls._km.shutdown_kernel(now=True)
            cls._km = None

    @classmethod
    def lock(cls) -> asyncio.Lock:
        """Get a lock to use for operations.

        Returns
        -------
        asyncio.Lock
            A lock.
        """
        return cls._lock



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Execution engines."""

from .base import Engine
from .factory import SUPPORTED_EXTS, make_engine
from .kernel import KernelManager

__all__ = [
    "SUPPORTED_EXTS",
    "Engine",
    "make_engine",
    "KernelManager",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Directories to use for user uploads and flow outputs."""

import base64
import builtins
import os
import sys
from functools import lru_cache
from pathlib import Path


@lru_cache
def is_frozen() -> bool:
    """Check if we are inside a compiled app.

    Returns
    -------
    bool
        True if detected frozen.
    """
    try:
        compiled = getattr(builtins, "__compiled__", False)
    except Exception:  # pylint: disable=broad-exception-caught
        compiled = False

    return bool(
        getattr(sys, "frozen", False)  # PyInstaller, cx_Freeze
        or hasattr(sys, "_MEIPASS")  # PyInstaller
        or compiled  # Nuitka/Cython
    )


@lru_cache
def is_installed_package() -> bool:
    """Check if running from an installed package (not editable/dev mode).

    Returns
    -------
    bool
        True if detected as installed, False otherwise.
    """
    try:
        # Check if running from site-packages
        module_path = Path(__file__).resolve()
        site_packages = any("site-packages" in str(p) for p in sys.path)
        return site_packages and "site-packages" in str(module_path)
    except Exception:  # pylint: disable=broad-exception-caught
        return False


@lru_cache
def get_root_dir(user_id: str = "default") -> Path:
    """Get the root directory for flows, user uploads, etc.

    Parameters
    ----------
    user_id : str
        The user ID (if using multi-user mode)

    Returns
    -------
    Path
        The root waldiez directory
    """
    if user_id == "default" and (
        is_frozen() or is_installed_package()
    ):  # pragma: no cover
        # root_dir = Path.home() / "waldiez" / "workspace"
        root_dir = Path.home()
        if root_dir.name.lower() != "waldiez":
            root_dir = root_dir / "waldiez"
        root_dir = root_dir / "workspace"
        root_dir.mkdir(parents=True, exist_ok=True)
        return root_dir
    files_root = Path(__file__).parent.parent / "files"
    root_dir = files_root / user_id
    is_testing = (
        os.environ.get("WALDIEZ_STUDIO_TESTING", "False").lower() == "true"
    )
    if is_testing:
        root_dir = (
            Path(os.environ.get("WALDIEZ_STUDIO_ROOT_DIR", str(files_root)))
            / user_id
        )
    root_dir.mkdir(parents=True, exist_ok=True)
    return root_dir.resolve(True)


def get_static_dir() -> Path:
    """Get the static directory.

    Returns
    -------
    Path
        The static directory
    """
    root_dir = Path(__file__).parent.parent / "static"
    is_testing = (
        os.environ.get("WALDIEZ_STUDIO_TESTING", "False").lower() == "true"
    )
    if is_testing:
        root_dir = Path(
            os.environ.get("WALDIEZ_STUDIO_ROOT_DIR", str(root_dir))
        )
    return root_dir


def path_to_id(path: Path) -> str:
    """Convert a Path object to a Base64-encoded string identifier.

    Parameters
    ----------
    path : Path
        The Path object to convert.

    Returns
    -------
    str
        Base64-encoded string identifier.
    """
    return base64.urlsafe_b64encode(path.as_posix().encode()).decode()


def id_to_path(identifier: str) -> Path:
    """Convert a Base64-encoded string identifier back to a Path object.

    Parameters
    ----------
    identifier : str
        The Base64-encoded string identifier.

    Returns
    -------
    Path
        Reconstructed Path object.
    """
    return Path(base64.urlsafe_b64decode(identifier).decode())



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Models."""

from .flow import SaveFlowRequest
from .workspace import (
    MessageResponse,
    PathItem,
    PathItemCreateRequest,
    PathItemListResponse,
    PathItemRenameRequest,
    PathItemType,
    SaveRequest,
)

__all__ = [
    "PathItem",
    "PathItemType",
    "PathItemCreateRequest",
    "PathItemRenameRequest",
    "PathItemListResponse",
    "MessageResponse",
    "SaveRequest",
    "SaveFlowRequest",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pylint: disable=broad-exception-caught
# pyright: reportCallInDefaultInitializer=false
"""Websocket routes."""

import contextlib
import logging
from pathlib import Path

from fastapi import APIRouter, Depends, WebSocket, WebSocketDisconnect

from waldiez_studio.engines import SUPPORTED_EXTS, Engine, make_engine
from waldiez_studio.utils.paths import path_to_id
from waldiez_studio.utils.task_runner import TaskRunner

from .common import check_path, get_root_directory

router = APIRouter()

LOG = logging.getLogger(__name__)


# pylint: disable=unused-argument
# noinspection PyBroadException
@router.websocket("/ws")
async def websocket_endpoint(
    websocket: WebSocket,
    path: str,
    root_dir: Path = Depends(get_root_directory),
) -> None:
    """Websocket endpoint.

    Parameters
    ----------
    websocket : WebSocket
        The websocket.
    path : str
        The path to the flow.
    root_dir : Path
        The root directory of the workspace
    """
    try:
        file_path = check_path(
            path,
            root_dir,
            path_type="File",
            must_exist=True,
            must_be_file=True,
            must_be_dir=False,
            must_have_extension=SUPPORTED_EXTS,
        )
    except BaseException as exc:
        LOG.error("Error: %s", exc)
        await websocket.close(code=1008, reason="Invalid path")
        return
    engine: Engine | None = None
    error: Exception | None = None
    # pylint: disable=too-many-try-statements
    try:
        await websocket.accept()
        engine = await make_engine(
            file_path=file_path, root_dir=root_dir, websocket=websocket
        )
        task_id = path_to_id(file_path)
        runner = TaskRunner(
            task_id=task_id,
            websocket=websocket,
            engine=engine,
        )
        await runner.listen()
    except WebSocketDisconnect as exc:
        error = exc
        LOG.info("WebSocket disconnected: %s", file_path)
    except Exception as exc:  # pylint: disable=broad-exception-caught
        error = exc  # pylint: disable=redefined-variable-type
        LOG.warning("WebSocket error for %s: %s", file_path, exc)
    finally:
        # if engine:
        #     with contextlib.suppress(Exception):
        #         await engine.shutdown()
        with contextlib.suppress(Exception):
            code = 1000 if not error else 1006
            reason = str(error) if error else None
            await websocket.close(code=code, reason=reason)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Models for the workspace API."""
# mypy: disable-error-code="unused-ignore,assignment"

from pydantic import BaseModel

try:
    from typing import Annotated, Literal  # noqa
except ImportError:
    from typing_extensions import Annotated, Literal  # noqa


PathItemType = Literal["file", "folder"]


# pylint: disable=too-few-public-methods
class PathItem(BaseModel):
    """Represents a file or folder in the file system."""

    name: str
    path: str
    type: PathItemType


class PathItemCreateRequest(BaseModel):
    """Request for creating a new file or folder."""

    type: PathItemType
    parent: Annotated[str | None, "The parent folder path."] = None


class PathItemRenameRequest(BaseModel):
    """Request for renaming a file or folder."""

    old_path: str
    new_path: str


class MessageResponse(BaseModel):
    """Standard response for successful operations."""

    message: str


class PathItemListResponse(BaseModel):
    """Response for listing files and folders."""

    items: list[PathItem]


class SaveRequest(BaseModel):
    """Save file contents request."""

    path: str
    content: str



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Configuration module for Waldiez Studio."""

from .lib import get_default_domain_name, get_default_host, get_default_port
from .settings import Settings, get_settings

__all__ = [
    "Settings",
    "get_settings",
    "get_default_domain_name",
    "get_default_host",
    "get_default_port",
]



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Logging configuration module."""

from typing import Any

import uvicorn.config


# fmt: off
def get_logging_config(log_level: str) -> dict[str, Any]:
    """Get logging config dict.

    Parameters
    ----------
    log_level : str
        The log level

    Returns
    -------
    dict[str, Any]
        The logging config dict
    """
    # skip spamming logs from these modules
    modules_to_have_level_info = [
        "watchgod",
        "watchfiles",
        "httpcore",
        "httpx",
    ]
    logging_config = uvicorn.config.LOGGING_CONFIG
    logging_config["formatters"]["default"]["fmt"] = (
        "%(levelprefix)s [%(name)s:%(filename)s:%(lineno)d] %(message)s"
    )
    logging_config["loggers"]["uvicorn"]["level"] = log_level
    logging_config["loggers"]["uvicorn.error"]["level"] = log_level
    logging_config["loggers"]["uvicorn.access"]["level"] = log_level
    httpx_logger = logging_config["loggers"].get("httpx", {})
    logging_config["loggers"]["httpx"] = httpx_logger
    logging_config["loggers"]["httpx"]["level"] = log_level
    logging_config["loggers"]["httpx"]["handlers"] = ["default"]
    http_core_logger = logging_config["loggers"].get("httpcore", {})
    logging_config["loggers"]["httpcore"] = http_core_logger
    logging_config["loggers"]["httpcore"]["handlers"] = ["default"]
    logging_config["loggers"]["httpcore"]["level"] = log_level
    for module in modules_to_have_level_info:
        logging_config["loggers"][module] = {
            "handlers": ["default"],
            "level": "INFO",
            "propagate": False,
        }
    logging_config["loggers"][""] = {
        "handlers": ["default"],
        "level": log_level,
        "propagate": False,
    }
    return logging_config
# fmt: on



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pyright: reportCallInDefaultInitializer=false,reportUnknownMemberType=false

"""Workspace API routes."""

import logging
import shutil
import tempfile
from pathlib import Path
from typing import Any

import aiofiles
import aiosqlite
import puremagic
from aiofiles import os
from fastapi import (
    APIRouter,
    BackgroundTasks,
    Depends,
    File,
    Form,
    HTTPException,
    Query,
    UploadFile,
)
from fastapi.responses import FileResponse, ORJSONResponse, Response

from waldiez_studio.models import (
    MessageResponse,
    PathItem,
    PathItemCreateRequest,
    PathItemListResponse,
    PathItemRenameRequest,
    PathItemType,
    SaveRequest,
)
from waldiez_studio.utils.sync import sync_to_async

from .common import (
    ALLOWED_EXTENSIONS,
    TEXTUAL_EXTS,
    check_path,
    get_new_file_name,
    get_new_folder_name,
    get_root_directory,
)

api = APIRouter()
LOG = logging.getLogger(__name__)

CHUNK_SIZE = 1024 * 1024  # 1 MB chunks
MAX_FILE_SIZE = 50 * 1024 * 1024  # 50 MB maximum file size
MAX_FILE_SIZE_MB = f"{MAX_FILE_SIZE / 1024 / 1024:.1f}"
SQLITE_EXTS = {".db", ".sqlite", ".sqlite3"}


@api.get("/workspace", response_model=PathItemListResponse)
async def list_items(
    parent: str | None = Query(
        default=None,
        description=(
            "The directory path to list. "
            "If not provided, the root directory will be listed."
        ),
    ),
    root_dir: Path = Depends(get_root_directory),
) -> PathItemListResponse:
    """
    List files and folders in a directory.

    Parameters
    ----------
    parent : Optional[str]
        The directory path to list.
        If not provided, the root directory will be listed.
    root_dir : Path
        The root directory to use.

    Returns
    -------
    PathItemListResponse
        A list of files and folders in the specified directory.

    Raises
    ------
    HTTPException
        If the directory does not exist.
    """
    target_dir = (
        root_dir
        if not parent
        else check_path(parent, root_dir=root_dir, must_be_dir=True)
    )

    def sync_list_items(directory: Path) -> list[PathItem]:
        """List files and folders in a given directory.

        Parameters
        ----------
        directory : Path
            The directory to list.

        Returns
        -------
        list[PathItem]
            A list of files and folders in the specified directory.
        """
        items = [
            PathItem(
                name=item.name,
                path=str(item.relative_to(root_dir)),
                type="folder" if item.is_dir() else "file",
            )
            for item in directory.iterdir()
        ]
        return sorted(items, key=lambda p: (p.type == "file", p.name.lower()))

    entries = await sync_to_async(sync_list_items)(target_dir)
    return PathItemListResponse(items=entries)


@api.post(
    "/workspace",
    response_model=PathItem,
    responses={
        "200": {"description": "The newly created file or folder"},
        "400": {"description": "Error: Invalid path"},
        "404": {"description": "Error: Path or parent directory not found"},
    },
)
async def create_file_or_folder(
    data: PathItemCreateRequest,
    root_dir: Path = Depends(get_root_directory),
) -> PathItem:
    """Create a new file or folder in the root directory.

    Parameters
    ----------
    data : PathItemCreateRequest
        The type of the item to create ("file" or "folder")
        and the parent folder path.
    root_dir : Path
        The root directory to use.

    Returns
    -------
    PathItem
        The newly created file or folder.

    Raises
    ------
    HTTPException
        If the file or folder already exists.
    """
    target_dir = (
        root_dir
        if not data.parent
        else check_path(data.parent, root_dir=root_dir, must_be_dir=True)
    )

    new_name = (
        get_new_folder_name(target_dir, "New Folder")
        if data.type == "folder"
        else get_new_file_name(target_dir, "Untitled.waldiez")
    )
    full_path = target_dir / new_name
    if full_path.exists():
        raise HTTPException(
            status_code=400, detail="Error: File or folder already exists"
        )
    return await create_new_path_item(
        full_path, item_type=data.type, root_dir=root_dir
    )


@api.post(
    "/workspace/rename",
    response_model=PathItem,
    responses={
        "200": {"description": "The renamed file or folder"},
        "400": {"description": "Error: Invalid path"},
        "404": {"description": "Error: Source file or folder does not exist"},
        "500": {"description": "Failed to rename file or folder"},
    },
)
async def rename_file_or_folder(
    data: PathItemRenameRequest,
    root_dir: Path = Depends(get_root_directory),
) -> PathItem:
    """
    Rename a file or folder.

    Parameters
    ----------
    data : PathItemRenameRequest
        The old and new paths for the file or folder.
    root_dir : Path
        The root directory to use.

    Returns
    -------
    PathItem
        The renamed file or folder.

    Raises
    ------
    HTTPException
        If the old path does not exist,
        the new path already exists,
        or renaming fails.
    """
    old_path = data.old_path
    new_path = data.new_path
    old_target = check_path(old_path, root_dir=root_dir, must_exist=True)
    new_target = check_path(
        new_path, root_dir=root_dir, must_exist=False, must_not_exist=True
    )
    try:
        old_target.rename(new_target)
    except BaseException as error:
        raise HTTPException(
            status_code=500, detail="Failed to rename file or folder"
        ) from error

    return PathItem(
        name=new_target.name,
        path=str(new_target.relative_to(root_dir)),
        type="folder" if new_target.is_dir() else "file",
    )


@api.post(
    "/workspace/save",
    response_model=MessageResponse,
    responses={
        "200": {"description": "The file was saved"},
        "400": {"description": "Error: Invalid request"},
        "404": {"description": "Error: File not found"},
        "415": {"description": "Error: Unsupported file type"},
        "500": {"description": "Failed to save file"},
    },
)
async def save_text_file(
    payload: SaveRequest,
    root_dir: Path = Depends(get_root_directory),
) -> MessageResponse:
    """
    Save a textual file (UTF-8).

    Parameters
    ----------
    payload : SaveRequest
        The file contents to store.
    root_dir : Path
        The root directory to use.

    Returns
    -------
    MessageResponse
        A success message.

    Raises
    ------
    HTTPException
        If the path does not exist,
        the file type is not supported
        or saving the file fails.
    """
    try:
        file_path = check_path(
            payload.path,
            root_dir,
            path_type="File",
            must_exist=True,
            must_be_file=True,
            must_be_dir=False,
        )
    except BaseException as exc:
        raise HTTPException(status_code=404, detail="File not found") from exc

    ext = file_path.suffix.lower()
    mime = ALLOWED_EXTENSIONS.get(ext)
    if not mime or ext not in TEXTUAL_EXTS:
        raise HTTPException(status_code=415, detail="Unsupported file type")

    try:
        # write text (replace invalids if ever present in content)
        file_path.write_text(payload.content, encoding="utf-8", errors="strict")
    except BaseException as exc:
        raise HTTPException(
            status_code=500, detail="Failed to save file"
        ) from exc

    return MessageResponse(message="Saved")


@api.post(
    "/workspace/upload",
    response_model=PathItem,
    responses={
        "200": {"description": "The uploaded file"},
        "400": {"description": "Invalid request"},
        "404": {"description": "Error: Directory not found"},
        "500": {"description": "Failed to upload file"},
    },
)
async def upload_file(
    path: str = Form(...),
    file: UploadFile = File(...),
    root_dir: Path = Depends(get_root_directory),
) -> PathItem:
    """Upload a file in chunks.

    Parameters
    ----------
    path : str
        The directory path where the file will be uploaded.
    file : UploadFile
        The file to upload.
    root_dir : Path
        The root directory to use.

    Returns
    -------
    PathItem
        The uploaded file.

    Raises
    ------
    HTTPException
        If the file exceeds the maximum size,
        if the already exists or there is an upload error.
    """
    if not file.filename:
        raise HTTPException(status_code=400, detail="Error: Invalid file name")
    target_dir = check_path(
        path, root_dir=root_dir, must_exist=True, must_be_dir=True
    )
    full_path = target_dir / file.filename
    if full_path.exists():
        raise HTTPException(
            status_code=400, detail="Error: File already exists"
        )
    file_size = 0
    # pylint: disable=too-many-try-statements
    try:
        async with aiofiles.open(full_path, "wb") as f:
            while chunk := await file.read(CHUNK_SIZE):
                file_size += len(chunk)
                if file_size > MAX_FILE_SIZE:
                    raise HTTPException(
                        status_code=400,
                        detail="Error: File exceeds maximum size",
                    )
                await f.write(chunk)
    except HTTPException as error:
        full_path.unlink(missing_ok=True)
        raise error
    except BaseException as error:
        full_path.unlink(missing_ok=True)
        raise HTTPException(
            status_code=500, detail="Failed to upload file"
        ) from error

    return PathItem(
        name=full_path.name,
        path=str(full_path.relative_to(root_dir)),
        type="file",
    )


@api.get(
    "/workspace/download",
    responses={
        "200": {"description": "The file was downloaded"},
        "400": {"description": "Error: Invalid path"},
        "404": {"description": "Error: File or folder not found"},
        "500": {"description": "Failed to download file or folder"},
    },
)
async def download_file_or_folder(
    path: str,
    background_tasks: BackgroundTasks,
    root_dir: Path = Depends(get_root_directory),
) -> FileResponse:
    """Download a file or folder.

    Parameters
    ----------
    path : str
        The path to the file or folder to download.
    background_tasks : BackgroundTasks
        The background tasks to use for deleting temporary files.
    root_dir : Path
        The root directory to use.

    Returns
    -------
    FileResponse
        The file or folder to download.

    Raises
    ------
    HTTPException
        If the file or folder does not exist or the download fails.
    """
    # make a zip if folder
    target_path = check_path(
        path, root_dir=root_dir, must_exist=True, must_be_dir=False
    )
    if target_path.is_file():
        return FileResponse(target_path, filename=target_path.name)
    return await sync_to_async(download_zip_sync)(
        target_path, root_dir, background_tasks
    )


@api.get(
    "/workspace/get",
    responses={
        "200": {"description": "The file or folder was deleted"},
        "400": {"description": "Error: Invalid path"},
        "404": {"description": "Error: File or folder not found"},
        "415": {"description": "Error: Unsupported file type"},
        "500": {"description": "Failed to delete file or folder"},
    },
)
async def get_file(
    path: str,
    root_dir: Path = Depends(get_root_directory),
) -> Response:
    """Get a file.

    Parameters
    ----------
    path : str
        The path to the file to get
    root_dir : Path
        The root directory to use.

    Returns
    -------
    Response
        A response with the file contents or the file itself

    Raises
    ------
    HTTPException
        If the request is invalid.
    """
    try:
        file_path = check_path(
            path,
            root_dir,
            path_type="File",
            must_exist=True,
            must_be_file=True,
            must_be_dir=False,
        )
    except BaseException as exc:
        LOG.error("Error: %s", exc)
        raise HTTPException(status_code=404, detail="Invalid request") from exc

    ext = file_path.suffix.lower()
    mime = ALLOWED_EXTENSIONS.get(ext)
    if not mime:
        try:
            guessed = puremagic.from_file(str(file_path))
            ext = _norm_ext(guessed)
            mime = ALLOWED_EXTENSIONS.get(ext)
        except BaseException as exc:
            raise HTTPException(
                status_code=415, detail="Unsupported file type"
            ) from exc

    if not mime:
        raise HTTPException(status_code=415, detail="Unsupported file type")

    if ext in TEXTUAL_EXTS:
        content = file_path.read_text(encoding="utf-8", errors="replace")
        return ORJSONResponse(
            {
                "path": str(file_path.relative_to(root_dir)),
                "mime": mime,
                "content": content,
            }
        )

    headers = {"Content-Disposition": f'inline; filename="{file_path.name}"'}
    return FileResponse(
        file_path,
        filename=file_path.name,
        media_type=mime,
        headers=headers,
    )


@api.delete(
    "/workspace",
    response_model=MessageResponse,
    responses={
        "200": {"description": "The file or folder was deleted"},
        "400": {"description": "Error: Invalid path"},
        "404": {"description": "Error: File or folder not found"},
        "500": {"description": "Failed to delete file or folder"},
    },
)
async def delete_file_or_folder(
    path: str, root_dir: Path = Depends(get_root_directory)
) -> MessageResponse:
    """Delete a file or folder.

    Parameters
    ----------
    path : str
        The path to the file or folder to delete.
    root_dir : Path
        The root directory to use.

    Returns
    -------
    MessageResponse
        A success message.

    Raises
    ------
    HTTPException
        If the file or folder does not exist.
    """
    target_path = check_path(
        path, root_dir=root_dir, must_exist=True, must_be_dir=False
    )
    thing = "Folder" if target_path.is_dir() else "File"
    try:
        if target_path.is_dir():
            rmtree = sync_to_async(shutil.rmtree)
            # pylint: disable=line-too-long
            # fmt: off
            await rmtree(target_path, ignore_errors=True)  # type: ignore[unused-ignore,call-arg]  # noqa: E501
            # fmt: on
        else:
            await os.remove(target_path)
    except BaseException as error:
        raise HTTPException(
            status_code=500, detail=f"Failed to delete {thing.lower()}"
        ) from error

    return MessageResponse(message=f"{thing} deleted successfully")


@api.get("/workspace/sqlite-tables")
async def sqlite_tables(
    path: str,
    root_dir: Path = Depends(get_root_directory),
) -> dict[str, Any]:
    """Get sqlite tables.

    Parameters
    ----------
    path : str
        The path to the sqlite db.
    root_dir : Path
        The root directory to use.

    Returns
    -------
    dict[str, Any]
        The tables in the db file.
    """
    file_path = check_path(path, root_dir, must_be_file=True)
    _ensure_sqlite(file_path)
    q = (
        "SELECT name FROM sqlite_master WHERE type='table' "
        "AND name NOT LIKE 'sqlite_%' ORDER BY name"
    )
    async with aiosqlite.connect(f"file:{file_path}?mode=ro", uri=True) as db:
        async with db.execute(q) as cur:
            tables = [r[0] async for r in cur]
    return {"tables": tables}


@api.get("/workspace/sqlite-rows")
async def sqlite_rows(  # pylint: disable=too-many-locals
    path: str,
    table: str = Query(..., description="Table name (must exist)"),
    limit: int = Query(50, ge=1, le=500),
    offset: int = Query(0, ge=0),
    order_by: str | None = Query(None, description="Column name to order by"),
    order_dir: str | None = Query(None, pattern="^(?i)(asc|desc)$"),
    search: str | None = Query(None, description="Simple substring filter"),
    root_dir: Path = Depends(get_root_directory),
) -> dict[str, Any]:
    """Get a table's rows.

    Parameters
    ----------
    path : str
        The path of the sqlite db file.
    table : str
        The table name.
    limit : int
        Optional limit for the number of rows. Defaults to 50
    offset : int
        Optional offset to use for fetching the rows. Defaults to 0
    order_by : str | None
        Optional ordering for the rows.
    order_dir : str
        Ordering direction (arc/desc).
    search : str
        Optional substring filter to search for.
    root_dir : Path
        The root directory to use.

    Returns
    -------
    dict[str, Any]
        The query results.

    Raises
    ------
    HTTPException
        If the table is not found.
    """
    file_path = check_path(path, root_dir, must_be_file=True)
    _ensure_sqlite(file_path)

    async with aiosqlite.connect(f"file:{file_path}?mode=ro", uri=True) as db:
        # verify table
        async with db.execute(
            "SELECT name FROM sqlite_master WHERE type='table' AND name=?",
            (table,),
        ) as cur:
            if not await cur.fetchone():
                raise HTTPException(status_code=404, detail="Table not found")

        # columns + types
        async with db.execute(
            f"PRAGMA table_info({_quote_ident(table)})"
        ) as cur:
            info: list[aiosqlite.Row] = [row async for row in cur]
        columns: list[str] = [r[1] for r in info]
        types = {r[1]: (r[2] or "").upper() for r in info}  # name -> decl type

        # build WHERE for search (LIKE across text-ish columns)
        where_sql = ""
        where_args: list[object] = []
        if search:
            like = f"%{search}%"
            text_cols = [
                column
                for column in columns
                if "CHAR" in types[column]
                # cspell: disable-next-line
                or "CLOB" in types[column]
                or "TEXT" in types[column]
                or types[column] == ""
            ]
            if text_cols:
                clauses = [f"{_quote_ident(c)} LIKE ?" for c in text_cols]
                where_sql = "WHERE " + " OR ".join(clauses)
                where_args = [like] * len(text_cols)

        # order by (validate column)
        order_sql = ""
        if order_by and order_by in columns:
            dir_sql = "DESC" if (order_dir or "").lower() == "desc" else "ASC"
            order_sql = f"ORDER BY {_quote_ident(order_by)} {dir_sql}"

        # total
        q = (
            "SELECT COUNT(*) FROM "  # nosemgrep # nosec
            f"{_quote_ident(table)} {where_sql}"
        )
        async with db.execute(
            q,
            where_args,
        ) as cur:
            total: aiosqlite.Row | None = await cur.fetchone()

        # rows
        q = (
            f"SELECT * FROM {_quote_ident(table)} "  # nosemgrep # nosec
            f"{where_sql} {order_sql} LIMIT ? OFFSET ?"
        )
        async with db.execute(
            q,
            [*where_args, limit, offset],
        ) as cur:
            rows = [r async for r in cur]

    return {
        "table": table,
        "columns": columns,
        "rows": rows,
        "total": total,
        "limit": limit,
        "offset": offset,
    }


async def create_new_path_item(
    full_path: Path, item_type: PathItemType, root_dir: Path
) -> PathItem:
    """Create a new file or folder.

    Parameters
    ----------
    full_path : Path
        The full path where the file or folder will be created.
    item_type : PathItemType
        The type of the item to create ("file" or "folder").
    root_dir : Path
        The root directory to use.

    Returns
    -------
    PathItem
        The newly created file or folder.
    Raises
    ------
    HTTPException
        If the item type is invalid or creation fails.
    """
    try:
        if item_type == "folder":
            full_path.mkdir(exist_ok=False)
        elif item_type == "file":
            async with aiofiles.open(full_path, mode="w") as f:
                await f.write('{"type": "flow"}')
    except BaseException as error:
        raise HTTPException(
            status_code=500, detail=f"Error: Failed to create {item_type}"
        ) from error
    return PathItem(
        name=full_path.name,
        path=str(full_path.relative_to(root_dir)),
        type=item_type,
    )


def download_zip_sync(
    target_path: Path, root_dir: Path, background_tasks: BackgroundTasks
) -> FileResponse:
    """Download a folder as a zip file.

    Parameters
    ----------
    target_path : Path
        The path to the folder to download.
    root_dir : Path
        The root directory to use.
    background_tasks : BackgroundTasks
        The background tasks to use for deleting temporary files.

    Returns
    -------
    FileResponse
        The zip file to download.

    Raises
    ------
    HTTPException
        If the download (archiving) fails.
    """
    # pylint: disable=too-many-try-statements
    try:
        # Create a temporary directory
        temp_dir = tempfile.mkdtemp()
        temp_dir_path = Path(temp_dir)
        zip_path = temp_dir_path / f"{target_path.name}.zip"

        # Create the archive
        archive = shutil.make_archive(
            base_name=str(zip_path.with_suffix("")),
            format="zip",
            root_dir=str(root_dir),
            base_dir=target_path.name,
        )

        # Add cleanup task for the temporary directory
        def cleanup() -> None:
            """Clean up the temporary directory."""
            shutil.rmtree(temp_dir, ignore_errors=True)

        background_tasks.add_task(cleanup)

        # Return the archive as a FileResponse
        return FileResponse(
            archive, filename=zip_path.name, background=background_tasks
        )

    except BaseException as error:
        raise HTTPException(
            status_code=500, detail="Error: Failed to download folder"
        ) from error


def _norm_ext(x: str) -> str:
    x = x.strip().lower()
    return x if x.startswith(".") else f".{x}"


def _ensure_sqlite(path: Path) -> None:
    if path.suffix.lower() not in SQLITE_EXTS:
        raise HTTPException(status_code=415, detail="Not a SQLite file")


def _quote_ident(name: str) -> str:
    # very simple quoting (escape double quotes)
    return '"' + name.replace('"', '""') + '"'



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Middleware for additional headers."""
# src/credits:
# https://github.com/fastapi/fastapi/discussions/8548#discussioncomment-5152780

import re
from collections import OrderedDict

from starlette.datastructures import MutableHeaders
from starlette.types import ASGIApp, Message, Receive, Scope, Send

CSP: dict[str, str | list[str]] = {
    "default-src": "'none'",
    "style-src": ["'self'", "'unsafe-inline'", "https://cdn.jsdelivr.net/npm/"],
    "script-src": [
        "'self'",
        "'wasm-unsafe-eval'",
        "https://drag-drop-touch-js.github.io/",
        "https://cdn.jsdelivr.net/npm/",
    ],
    "img-src": ["*", "blob:", "data:"],
    "worker-src": ["'self'", "blob:"],
    "connect-src": "*",
    "font-src": ["'self'", "https://cdn.jsdelivr.net/npm/", "data:"],
    "manifest-src": "'self'",
    "media-src": ["'self'", "data:"],
}


def parse_policy(policy: dict[str, str | list[str]] | str) -> str:
    """Parse a given policy dict to string.

    Parameters
    ----------
    policy : dict[str, str | list[str]] | str
        The policy dict or string

    Returns
    -------
    str
        The parsed policy string
    """
    if isinstance(policy, str):
        # parse the string into a policy dict
        policy_string = policy
        policy = OrderedDict()

        for policy_part in policy_string.split(";"):
            policy_parts = policy_part.strip().split(" ")
            policy[policy_parts[0]] = " ".join(policy_parts[1:])

    policies: list[str] = []
    for section, content in policy.items():
        if not isinstance(content, str):
            content = " ".join(content)
        policy_part = f"{section} {content}"

        policies.append(policy_part)

    parsed_policy = "; ".join(policies)

    return parsed_policy


# pylint: disable=too-few-public-methods
class ExtraHeadersMiddleware:
    """Add extra headers to all responses."""

    def __init__(
        self,
        app: ASGIApp,
        exclude_patterns: list[str] | None = None,
        csp: bool = True,
        force_ssl: bool = True,
        max_age: int = 31556926,
    ):
        """Init SecurityHeadersMiddleware.

        Parameters
        ----------
        app: ASGIApp
            The ASGI Application
        exclude_patterns: list[str], optional
            List of regex patterns to exclude from adding headers
        csp: bool
            Whether to add a Content-Security-Policy header
        force_ssl: bool
            Whether to add a Strict-Transport-Security header
        max_age: int
            The max age for the Strict-Transport-Security header
        """
        self.app = app
        self.csp = csp
        self.force_ssl = force_ssl
        self.max_age = max_age
        self.exclude_patterns = [
            re.compile(p) for p in (exclude_patterns or [])
        ]
        self._policy = parse_policy(CSP)

    async def __call__(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None:
        """Call the middleware.

        Parameters
        ----------
        scope : Scope
            The ASGI scope
        receive : Receive
            The ASGI receive channel
        send : Send
            The ASGI send channel
        """
        scope_path = scope.get("path", "")
        scope_type = scope.get("type")
        skip = scope_type != "http"
        # skip = scope.get("type", "") != "http"
        if not skip:
            skip = any(p.search(scope_path) for p in self.exclude_patterns)
        if skip:
            await self.app(scope, receive, send)
        else:

            async def send_wrapper(message: Message) -> None:
                if message["type"] == "http.response.start":
                    headers = MutableHeaders(scope=message)
                    additional_headers = {
                        "Cross-Origin-Opener-Policy": "same-origin",
                        "Referrer-Policy": "strict-origin-when-cross-origin",
                        "X-Content-Type-Options": "nosniff",
                        "X-Frame-Options": "DENY",
                        "X-XSS-Protection": "1; mode=block",
                    }
                    if self.csp:
                        additional_headers["Content-Security-Policy"] = (
                            self._policy
                        )
                    if self.force_ssl:
                        additional_headers["Strict-Transport-Security"] = (
                            f"max-age={self.max_age}; includeSubDomains"
                        )
                    headers.update(additional_headers)
                await send(message)

            await self.app(scope, receive, send_wrapper)



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

# pyright: reportUnusedParameter=false

"""Base abstract terminal session to inherit."""

import abc


class BaseSession(abc.ABC):
    """Base session class."""

    @abc.abstractmethod
    async def read(self, n: int = 4096) -> bytes:
        """Read.

        Parameters
        ----------
        n : int
            The number of rows to read.

        Returns
        -------
        bytes
            The read data.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def write(self, data: bytes) -> None:
        """Write.

        Parameters
        ----------
        data : bytes
            The data to pass.
        """
        raise NotImplementedError

    def resize(self, rows: int, cols: int) -> None:
        """Resize the terminal.

        Parameters
        ----------
        rows : int
            The new number of rows.
        cols : int
            The new number of columns.
        """

    def interrupt(self) -> None:
        """Interrupt the session."""

    def terminate(self) -> None:
        """Terminate the session."""

    # pylint: disable=no-self-use
    def is_alive(self) -> bool:
        """Check if the session is alive.

        Returns
        -------
        bool
            True if the session is alive, false otherwise.
        """
        return False

    def close(self) -> None:
        """Close the session."""



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.
"""Version information for waldiez_studio.

Dev:
    Do not edit this file.
    Use ../package.json to update the version instead.
    The built package version will be set automatically by Hatchling.
    See ../scripts/hooks/metadata.py for more information.
"""

__version__ = "0.6.1"



# SPDX-License-Identifier: Apache-2.0.
# Copyright (c) 2024 - 2025 Waldiez and contributors.

"""Custom middlewares for the waldiez_studio project."""

from .extra_headers import ExtraHeadersMiddleware

__all__ = ["ExtraHeadersMiddleware"]
